__all__ = ('ComponentMetadataMentionableSelect', )

from .entity_select_base import ComponentMetadataEntitySelectBase


class ComponentMetadataMentionableSelect(ComponentMetadataEntitySelectBase):
    """
    mentionable (User and role) select component metadata.
    
    Attributes
    ----------
    custom_id : `None`, `str`
        Custom identifier to detect which component was used by the user.
    
    default_values : `None`, `tuple` of ``EntitySelectDefaultValue``
        Entities presented in the select by default.
    
    enabled : `bool`
        Whether the component is enabled.
    
    max_values : `int
        The maximal amount of options to select.
    
    min_values : `int`
        The minimal amount of options to select.
    
    placeholder : `None`, `str`
        Placeholder text of the select.
    """
    __slots__ = ()



__all__ = ('OnboardingMode',)

from ...bases import Preinstance as P, PreinstancedBase


class OnboardingMode(PreinstancedBase):
    """
    Onboarding mode.
    
    Attributes
    ----------
    name : `str`
        The default name of the onboarding mode.
    value : `int`
        The discord side identifier value of the onboarding mode.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``OnboardingMode``) items
        Stores the predefined ``OnboardingMode``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The onboarding mode' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the onboarding mode.
    
    Every predefined onboarding mode can be accessed as class attribute as well:
    
    +-------------------------------+---------------------------+-------+
    | Class attribute name          | name                      | value |
    +===============================+===========================+=======+
    | default                       | default                   | 0     |
    +-------------------------------+---------------------------+-------+
    | advanced                      | advanced                  | 1     |
    +-------------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    
    default = P(0, 'default')
    advanced = P(1, 'advanced')



__all__ = ()

from datetime import datetime as DateTime
from functools import partial as partial_func
from warnings import warn

from ..voice_state.fields import (
    put_channel_id_into, put_deaf_into, put_mute_into, validate_channel_id, validate_deaf, validate_mute
)

from .fields import (
    put_nick_into, put_role_ids_into, put_timed_out_until_into, validate_nick, validate_role_ids,
    validate_timed_out_until, validate_timeout_duration
)
from .guild_profile import GUILD_PROFILE_AVATAR


def validate_timeout_duration_with_deprecation(timeout_duration):
    """
    Validates the given `timeout_duration`, but also accepts `DateTime` and `None` as before.
    
    Deprecated and will be removed in 2024 March.
    
    Parameters
    ----------
    timeout_duration : `None`, `DateTime`, `int`, `TimeDelta`, `float`
        Timeout duration.
    
    Returns
    -------
    timed_out_until : `None`, `DateTime`
    
    Raises
    ------
    TypeError
    """
    if (timeout_duration is None) or isinstance(timeout_duration, DateTime):
        warn(
            (
                f'`timeout_duration` parameter cannot be `None`, {DateTime.__name__}; '
                f'please use `timed_out_until` parameter instead. '
                f'This warning will be removed at 2024 March and `TypeError` will be propagated instead.'
            ),
            FutureWarning,
            stacklevel = 4,
        )
        
        return validate_timed_out_until(timeout_duration)
    
    return validate_timeout_duration(timeout_duration)


GUILD_PROFILE_FIELD_CONVERTERS = {
    'deaf': (validate_deaf, put_deaf_into),
    'mute': (validate_mute, put_mute_into),
    'nick': (validate_nick, put_nick_into),
    'role_ids': (validate_role_ids, put_role_ids_into),
    'roles': (validate_role_ids, put_role_ids_into),
    'timed_out_until': (validate_timed_out_until, put_timed_out_until_into),
    'timeout_duration': (validate_timeout_duration_with_deprecation, put_timed_out_until_into),
    'voice_channel': (validate_channel_id, put_channel_id_into),
    'voice_channel_id': (validate_channel_id, put_channel_id_into),
}

GUILD_PROFILE_SELF_FIELD_CONVERTERS = {
    'avatar': (
        partial_func(GUILD_PROFILE_AVATAR.validate_icon, allow_data = True),
        partial_func(GUILD_PROFILE_AVATAR.put_into, as_data = True),
    ),
    'nick': (validate_nick, put_nick_into),
}



__all__ = ()


from .fields import (
    put_default_channel_ids_into, put_enabled_into, put_mode_into, put_prompts_into, validate_default_channel_ids,
    validate_enabled, validate_mode, validate_prompts
)


ONBOARDING_FIELD_CONVERTERS = {
    'default_channel_ids': (validate_default_channel_ids, put_default_channel_ids_into),
    'enabled': (validate_enabled, put_enabled_into),
    'mode': (validate_mode, put_mode_into),
    'prompts': (validate_prompts, put_prompts_into),
}


# https://github.com/discord/discord-api-docs/pull/6479
def flatten_emoji_data_in_onboarding_screen_prompt_options(data):
    """
    Flattens an onboarding screen's prompts' options' emojis' datas.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Onboarding prompt data.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    prompt_datas = data.get('prompts', None)
    if (prompt_datas is not None):
        for prompt_data in prompt_datas:
            option_datas = prompt_data.get('options', None)
            if (option_datas is None):
                continue
            
            for option_data in option_datas:
                try:
                    emoji_data = option_data.pop('emoji')
                except KeyError:
                    continue
                
                if emoji_data is None:
                    option_data['emoji_name'] = None
                    continue
                
                for key, value in emoji_data.items():
                    option_data['emoji_' + key] = value

    return data



__all__ = ('PollLayout',)

from ...bases import Preinstance as P, PreinstancedBase


class PollLayout(PreinstancedBase):
    """
    Represents a poll's layout.
    
    Attributes
    ----------
    name : `str`
        The name of the poll layout.
    value : `int`
        The identifier value the poll layout.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``PollLayout``) items
        Stores the predefined ``PollLayout``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The poll layouts' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the poll layouts.
    
    Every predefined poll layout can be accessed as class attribute as well:
    
    +-----------------------+-------------------+-------+
    | Class attribute name  | Name              | Value |
    +=======================+===================+=======+
    | none                  | none              | 0     |
    +-----------------------+-------------------+-------+
    | default               | default           | 1     |
    +-----------------------+-------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    default = P(1, 'default')



__all__ = ()

from .fields import put_emoji_pack_collectible_into, validate_emoji_pack_collectible


GUILD_INVENTORY_SETTINGS_FIELD_CONVERTERS = {
    'emoji_pack_collectible': (validate_emoji_pack_collectible, put_emoji_pack_collectible_into),
}



__all__ = ()

from ...field_parsers import entity_id_parser_factory
from ...field_putters import entity_id_optional_putter_factory
from ...field_validators import entity_id_validator_factory, force_date_time_validator_factory
from ...utils import DISCORD_EPOCH_START, datetime_to_timestamp, timestamp_to_datetime

from ..message import MessageFlag
from ..message.fields import (
    parse_attachments as _parse_attachments, parse_content as _parse_content, parse_edited_at as _parse_edited_at,
    parse_embeds as _parse_embeds, parse_flags as _parse_flags, validate_attachments, validate_content,
    validate_edited_at, validate_embeds, validate_flags
)


# attachments

def parse_attachments(data):
    """
    Parses out attachments value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    attachments : `None | tuple<Attachment>`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_attachments(message_data)


def put_attachments_into(attachments, data, defaults):
    """
    Serializes the given attachments into the given data.
    
    Parameters
    ----------
    attachments : `None | tuple<Attachment>`
        The attachments to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (attachments is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if attachments is None:
            attachment_datas = []
        else:
            attachment_datas = [
                attachment.to_data(defaults = defaults, include_internals = True) for attachment in attachments
            ]
        
        message_data['attachments'] = attachment_datas
    
    return data


# content

def parse_content(data):
    """
    Parses out content value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    content : `None | str`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_content(message_data)


def put_content_into(content, data, defaults):
    """
    Serializes the given content into the given data.
    
    Parameters
    ----------
    content : `None | str`
        The content to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (content is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if content is None:
            content = ''
        
        message_data['content'] = content
    
    return data


# created_at

def parse_created_at(data):
    """
    Parses out created at value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    created_at : `None | DateTime`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return DISCORD_EPOCH_START
    
    created_at = message_data.get('timestamp', None)
    if (created_at is None):
        return DISCORD_EPOCH_START
    
    return timestamp_to_datetime(created_at)


def put_created_at_into(created_at, data, defaults):
    """
    Serializes the given created at value into the given data.
    
    Parameters
    ----------
    created_at : `None | DateTime`
        The created at value to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (created_at != DISCORD_EPOCH_START) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        message_data['timestamp'] = datetime_to_timestamp(created_at)
    
    return data


validate_created_at = force_date_time_validator_factory('created_at')

# edited_at

def parse_edited_at(data):
    """
    Parses out edited at value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    edited_at : `None | DateTime`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_edited_at(message_data)


def put_edited_at_into(edited_at, data, defaults):
    """
    Serializes the given created at value into the given data.
    
    Parameters
    ----------
    edited_at : `None | DateTime`
        The created at value to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (edited_at is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if (edited_at is None):
            timestamp = None
        else:
            timestamp = datetime_to_timestamp(edited_at)
        
        message_data['edited_timestamp'] = timestamp
    
    return data


# embeds

def parse_embeds(data):
    """
    Parses out embeds value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    embeds : `None | tuple<Embed>`
    """
    message_data = data.get('message', None)
    if message_data is None:
        return None
    
    return _parse_embeds(message_data)


def put_embeds_into(embeds, data, defaults):
    """
    Serializes the given embeds into the given data.
    
    Parameters
    ----------
    embeds : `None | tuple<Embed>`
        The embeds to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if (embeds is not None) or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        if embeds is None:
            embed_datas = []
        else:
            embed_datas = [embed.to_data(defaults = defaults) for embed in embeds]
        
        message_data['embeds'] = embed_datas
    
    return data


# flags

def parse_flags(data):
    """
    Parses out flags value from the given data.
    
    Parameters
    ----------
    data : `dict<str, object>`
        Data to parse from.
    
    Returns
    -------
    flags : ``MessageFlag``
    """
    message_data = data.get('message', None)
    if message_data is None:
        return MessageFlag(0)
    
    return _parse_flags(message_data)


def put_flags_into(flags, data, defaults):
    """
    Serializes the given flags into the given data.
    
    Parameters
    ----------
    flags : ``MessageFlag``
        The flags to serialize.
    data : `dict<str, object>`
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    
    Returns
    -------
    data : `dict<str, object>`
    """
    if flags or defaults:
        message_data = data.get('message', None)
        if message_data is None:
            message_data = {}
            data['message'] = message_data
        
        message_data['flags'] = int(flags)
    
    return data


# guild_id

parse_guild_id = entity_id_parser_factory('guild_id')
put_guild_id_into = entity_id_optional_putter_factory('guild_id')
validate_guild_id = entity_id_validator_factory('guild_id', NotImplemented, include = 'Guild')



__all__ = ('MessageReferenceType',)

from ...bases import Preinstance as P, PreinstancedBase


class MessageReferenceType(PreinstancedBase):
    """
    Represents an message reference's type.
    
    Attributes
    ----------
    name : `str`
        The name of the message reference type.
    value : `int`
        The Discord side identifier value of the message reference type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``MessageReferenceType``) items
        Stores the created message reference type instances. This container is accessed when translating a Discord
        message reference type's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The message reference types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the message reference types.
    
    Every predefined message reference type can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | reply                     | reply                     | 0     |
    +---------------------------+---------------------------+-------+
    | forward                   | forward                   | 1     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    reply = P(0, 'reply')
    forward = P(1, 'forward')



__all__ = ('Invite',)

from datetime import datetime as DateTime
from warnings import warn

from scarletio import export

from ...bases import DiscordEntity
from ...core import INVITES
from ...guild import NsfwLevel
from ...http import urls as module_urls
from ...precreate_helpers import process_precreate_parameters_and_raise_extra
from ...user import ClientUserBase, ZEROUSER

from .fields import (
    parse_approximate_online_count, parse_approximate_user_count, parse_channel, parse_code, parse_created_at,
    parse_flags, parse_guild, parse_inviter, parse_max_age, parse_max_uses, parse_target_application, parse_target_type,
    parse_target_user, parse_temporary, parse_type, parse_uses, put_approximate_online_count_into,
    put_approximate_user_count_into, put_channel_into, put_code_into, put_created_at_into, put_flags_into,
    put_guild_into, put_inviter_into, put_max_age_into, put_max_uses_into, put_target_application_id_into,
    put_target_application_into, put_target_type_into, put_target_user_id_into, put_target_user_into,
    put_temporary_into, put_type_into, put_uses_into, validate_approximate_online_count,
    validate_approximate_user_count, validate_channel, validate_code, validate_created_at, validate_flags,
    validate_guild, validate_inviter, validate_max_age, validate_max_uses, validate_target_application,
    validate_target_type, validate_target_user, validate_temporary, validate_type, validate_uses
)
from .flags import InviteFlag
from .preinstanced import InviteTargetType, InviteType


PRECREATE_FIELDS = {
    'approximate_user_count': ('approximate_user_count', validate_approximate_user_count),
    'approximate_online_count': ('approximate_online_count', validate_approximate_online_count),
    'channel': ('channel', validate_channel),
    'created_at': ('created_at', validate_created_at),
    'flags': ('flags', validate_flags),
    'guild': ('guild', validate_guild),
    'invite_type': ('type', validate_type),
    'inviter': ('inviter', validate_inviter),
    'max_age': ('max_age', validate_max_age),
    'max_uses': ('max_uses', validate_max_uses),
    'target_application': ('target_application', validate_target_application),
    'target_type': ('target_type', validate_target_type),
    'target_user': ('target_user', validate_target_user),
    'temporary': ('temporary', validate_temporary),
    'uses': ('uses', validate_uses),
}


@export
class Invite(DiscordEntity, immortal = True):
    """
    Represents a Discord Invite.
    
    Attributes
    ---------
    approximate_online_count : `int`
        The approximate amount of online users at the respective guild (or group channel).
        Defaults to `0`.
    
    approximate_user_count : `int`
        The approximate amount of users at the respective guild (or group channel).
        Defaults to `0`.
    
    channel : `None`, ``Channel``
        The channel where the invite redirects. If it is announcements or store channel, then the invite is a lurk
        invite. If channel data was not sent with the invite's, then this attribute is set as `None`.
    
    code : `str`
        The invite's unique identifier.
    
    created_at : `DateTime`
        When the invite was created. Defaults to Discord epoch.
    
    flags : ``InviteFlag``
        The invite's flags.
    
    guild : `None`, ``Guild``
        The guild the invite is for. If not included or if the invite's channel is a group channel, then set as
        `None`.
    
    inviter : ``ClientUserBase``
        The creator of the invite. If not included, then set as `ZEROUSER`.
    
    max_age : `None`, `int`
        The time in seconds after the invite will expire.
        Defaults to `None`.
        
        If the invite was created with max age as `0`, then this value will be negative instead of the expected `0`.
    
    max_uses : `None`, `int`
        How much times the invite can be used.
        Defaults to `None`.
        
        If the invite has no use limit, then this value is set as `0`.
    
    target_application : `None`, ``Application``
        The invite's target application.
    
    target_type : ``InviteTargetType``
        The invite's target type.
    
    target_user : `None`, ``ClientUserBase``
        The target of the invite if applicable.
    
    temporary : `bool`
        Whether this invite only grants temporary membership.
        
        When the user goes offline, they get kicked, except if they got a role meanwhile.
    
    type : ``InviteType``
        The invite's type.
    
    uses : `None`, `int`
        The amount how much times the invite was used.
        Defaults to `None`.
    """
    __slots__ = (
        'approximate_user_count', 'approximate_online_count', 'channel', 'code', 'created_at', 'flags', 'guild',
        'inviter', 'max_age', 'max_uses', 'target_application', 'target_type', 'target_user', 'temporary', 'type',
        'uses'
    )
    
    
    def __new__(
        cls,
        *, 
        flags = ...,
        max_age = ...,
        max_uses = ...,
        target_application = ...,
        target_type = ...,
        target_user = ...,
        temporary = ...,
    ):
        """
        Creates a partial invite with the given fields.
        
        Parameters
        ----------
        flags : ``InviteFlag``, `int`, Optional (Keyword only)
            The invite's flags.
        
        max_age : `None`, `int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        
        max_uses : `None`, `int`, Optional (Keyword only)
            How much times the invite can be used.
        
        target_application : `None`, ``Application``, Optional (Keyword only)
            The invite's target application.
        
        target_type : ``InviteTargetType``, `int`, Optional (Keyword only)
            The invite's target type.
        
        target_user : `None`, ``ClientUserBase``, Optional (Keyword only)
            The target of the invite if applicable.
        
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # flags
        if flags is ...:
            flags = InviteFlag()
        else:
            flags = validate_flags(flags)
        
        # max_age
        if max_age is ...:
            max_age = None
        else:
            max_age = validate_max_age(max_age)
        
        # max_uses
        if max_uses is ...:
            max_uses = None
        else:
            max_uses = validate_max_uses(max_uses)
        
        # target_application
        if target_application is ...:
            target_application = None
        else:
            target_application = validate_target_application(target_application)
        
        # target_type
        if target_type is ...:
            target_type = InviteTargetType.none
        else:
            target_type = validate_target_type(target_type)
        
        # target_user
        if target_user is ...:
            target_user = None
        else:
            target_user = validate_target_user(target_user)
        
        # temporary
        if temporary is ...:
            temporary = False
        else:
            temporary = validate_temporary(temporary)
        
        # Construct
        self = object.__new__(cls)
        
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.channel = None
        self.code = ''
        self.created_at = None
        self.flags = flags
        self.guild = None
        self.inviter = ZEROUSER
        self.max_age = max_age
        self.max_uses = max_uses
        self.target_application = target_application
        self.target_type = target_type
        self.target_user = target_user
        self.temporary = temporary
        self.type = InviteType.guild
        self.uses = None
        
        return self
        
    
    @classmethod
    def from_data(cls, data):
        """
        Creates an invite from the given invite data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Invite data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        code = parse_code(data)
        
        try:
            self = INVITES[code]
        except KeyError:
            self = object.__new__(cls)
            self.code = code
            self._set_attributes(data)
            INVITES[code] = self
        
        else:
            self._update_attributes(data)
        
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the invite to json serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields, like id-s should be present as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_flags_into(self.flags, data, defaults)
        put_max_age_into(self.max_age, data, defaults)
        put_max_uses_into(self.max_uses, data, defaults)
        put_target_type_into(self.target_type, data, defaults)
        put_temporary_into(self.temporary, data, defaults)
        
        if include_internals:
            put_approximate_user_count_into(self.approximate_user_count, data, defaults)
            put_approximate_online_count_into(self.approximate_online_count, data, defaults)
            put_channel_into(self.channel, data, defaults)
            put_code_into(self.code, data, defaults)
            put_created_at_into(self.created_at, data, defaults)
            put_guild_into(self.guild, data, defaults)
            put_inviter_into(self.inviter, data, defaults)
            put_target_application_into(self.target_application, data, defaults)
            put_target_user_into(self.target_user, data, defaults)
            put_type_into(self.type, data, defaults)
            put_uses_into(self.uses, data, defaults)
        
        else:
            put_target_application_id_into(self.target_application_id, data, defaults)
            put_target_user_id_into(self.target_user_id, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the representation of the invite."""
        return f'<{self.__class__.__name__} code = {self.code!r}>'
    
    
    def __hash__(self):
        """Returns the invite's code's hash."""
        code = self.code
        if code:
            return hash(code)
        return self._get_hash_partial()
    
    
    def _get_hash_partial(self):
        """
        Returns a partial invite's hash value.
        
        Returns
        -------
        hash_value : `int`
        """
        hash_value = 0
        
        # flags
        hash_value ^= self.flags
        
        # max_age
        hash_value ^= self.max_age << 4
        
        # max_uses
        hash_value ^= self.max_uses << 8
        
        # target_application
        target_application = self.target_application
        if (target_application is not None):
            hash_value ^= hash(target_application)
        
        # target_type
        hash_value ^= self.target_type.value << 12
        
        # target_user
        target_user = self.target_user
        if (target_user is not None):
            hash_value ^= hash(target_user)
        
        # temporary
        hash_value ^= self.temporary << 16
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two invites are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two invites are not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # flags
        if self.flags != other.flags:
            return False
        
        # max_age
        if self.max_age != other.max_age:
            return False
        
        # max_uses
        if self.max_uses != other.max_uses:
            return False
        
        # target_application
        if self.target_application is not other.target_application:
            return False
        
        # target_type
        if self.target_type is not other.target_type:
            return False
        
        # target_user
        if self.target_user is not other.target_user:
            return False
        
        # temporary
        if self.temporary != other.temporary:
            return False
        
        return True
    
    
    def _set_attributes(self, data):
        """
        Sets the invite's attributes (except code).
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Invite data.
        """
        self._set_attributes_common(data)
        self._set_counts_only(data)
    
    
    def _update_attributes(self, data):
        """
        Updates the invite with the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Invite data.
        """
        self._set_attributes_common(data)
        self._update_counts_only(data)
    
    
    def _update_attributes_partial(self, data):
        """
        Updates the invite's fields that are received with a partial data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Invite data.
        """
        self.channel = parse_channel(data)
        self.guild = parse_guild(data)
    
    
    def _set_attributes_common(self, data):
        """
        Sets the common attributes of the invite. Used both in ``._set_attributes`` and ``._update_attributes``.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Invite data.
        """
        self.channel = parse_channel(data)
        self.created_at = parse_created_at(data)
        self.flags = parse_flags(data)
        self.guild = parse_guild(data)
        self.inviter = parse_inviter(data)
        self.max_age = parse_max_age(data)
        self.max_uses = parse_max_uses(data)
        self.target_application = parse_target_application(data)
        self.target_type = parse_target_type(data)
        self.target_user = parse_target_user(data)
        self.temporary = parse_temporary(data)
        self.type = parse_type(data)
        self.uses = parse_uses(data)
    
    
    def _set_counts_only(self, data):
        """
        Sets the invite's counts.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received invite data.
        """
        guild = self.guild
        
        self.approximate_online_count = approximate_online_count = parse_approximate_online_count(data)
        if approximate_online_count and (guild is not None):
            guild.approximate_online_count = approximate_online_count
        
        self.approximate_user_count = approximate_user_count = parse_approximate_user_count(data)
        if approximate_user_count and (guild is not None):
            guild.approximate_user_count = approximate_user_count
    
    
    def _update_counts_only(self, data):
        """
        Updates the invite's counts if given.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received invite data.
        """
        guild = self.guild
        
        approximate_online_count = parse_approximate_online_count(data)
        if approximate_online_count:
            self.approximate_online_count = approximate_online_count
            if (guild is not None):
                guild.approximate_online_count = approximate_online_count
        
        approximate_user_count = parse_approximate_user_count(data)
        if approximate_user_count:
            self.approximate_user_count = approximate_user_count
            if (guild is not None):
                guild.approximate_user_count = approximate_user_count
    
    
    @classmethod
    def precreate(cls, code, **keyword_parameters):
        """
        Precreates an invite object with the given parameters.
        
        Parameters
        ----------
        code : `str`
            The invite's code.
        **keyword_parameters : keyword parameters
            Additional predefined attributes for the invite.
        
        Other Parameters
        ----------------
        approximate_online_count : `int`, Optional (Keyword only)
            The amount of online users at the respective guild (or group channel).
        
        approximate_user_count : `int`, Optional (Keyword only)
            The amount of users at the respective guild (or group channel).
        
        channel : `None`, ``Channel``, Optional (Keyword only)
            The channel where the invite redirects.
        
        created_at : `None`, `DateTime`, Optional (Keyword only)
            When the invite was created.
        
        flags : ``InviteFlag``, `int`, Optional (Keyword only)
            The invite's flags.
        
        guild : `None`, ``Guild``, Optional (Keyword only)
            The guild the invite is for.
        
        invite_type : `int`, ``InviteType``, Optional (Keyword only)
            The invite's type.
        
        inviter : ``ClientUserBase``, Optional (Keyword only)
            The creator of the invite.
        
        max_age : `None`, `int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        
        max_uses : `None`, `int`, Optional (Keyword only)
            How much times the invite can be used.
        
        target_application : `None`, ``Application``, Optional (Keyword only)
            The target application of the invite.
        
        target_type : ``InviteTargetType``, `int` Optional (Keyword only)
            The invite's target type.
        
        target_user : `None`, ``ClientUserBase``, Optional (Keyword only)
            The target user of the invite.
        
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        
        uses : `None`, `int`, Optional (Keyword only)
            The amount how much times the invite was used.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        code = validate_code(code)

        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        try:
            self = INVITES[code]
        except KeyError:
            self = cls._create_empty(code)
            INVITES[code] = self
        
        else:
            # Cannot detect if an invite is not partial -> do nothing
            pass
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, code):
        """
        Creates an empty invite with default attributes set.
        
        Parameters
        ----------
        code : `str`
            Unique identifier of the invite.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        
        self.approximate_online_count = 0
        self.approximate_user_count = 0
        self.channel = None
        self.code = code
        self.created_at = None
        self.flags = InviteFlag()
        self.guild = None
        self.inviter = ZEROUSER
        self.max_age = None
        self.max_uses = None
        self.target_application = None
        self.target_type = InviteTargetType.none
        self.target_user = ZEROUSER
        self.temporary = False
        self.type = InviteType.guild
        self.uses = None
        
        return self
    
    
    def copy(self):
        """
        Copies the invite returning a new partial one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        
        new.approximate_online_count = 0
        new.approximate_user_count = 0
        new.channel = None
        new.code = ''
        new.created_at = None
        new.flags = self.flags
        new.guild = None
        new.inviter = ZEROUSER
        new.max_age = self.max_age
        new.max_uses = self.max_uses
        new.target_application = self.target_application
        new.target_type = self.target_type
        new.target_user = self.target_user
        new.temporary = self.temporary
        new.type = InviteType.guild
        new.uses = None
        
        return new
    
    
    def copy_with(
        self,
        *, 
        flags = ...,
        max_age = ...,
        max_uses = ...,
        target_application = ...,
        target_type = ...,
        target_user = ...,
        temporary = ...,
    ):
        """
        Copies the invite with the given fields.
        
        Parameters
        ----------
        flags : ``InviteFlag``, `int`, Optional (Keyword only)
            The invite's flags.
        
        max_age : `None`, `int`, Optional (Keyword only)
            The time in seconds after the invite will expire.
        
        max_uses : `None`, `int`, Optional (Keyword only)
            How much times the invite can be used.
        
        target_application : `None`, ``Application``, Optional (Keyword only)
            The invite's target application.
        
        target_type : ``InviteTargetType``, `int`, Optional (Keyword only)
            The invite's target type.
        
        target_user : `None`, ``ClientUserBase``, Optional (Keyword only)
            The target of the invite if applicable.
        
        temporary : `bool`, Optional (Keyword only)
            Whether this invite only grants temporary membership.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # max_age
        if max_age is ...:
            max_age = self.max_age
        else:
            max_age = validate_max_age(max_age)
        
        # max_uses
        if max_uses is ...:
            max_uses = self.max_uses
        else:
            max_uses = validate_max_uses(max_uses)
        
        # target_application
        if target_application is ...:
            target_application = self.target_application
        else:
            target_application = validate_target_application(target_application)
        
        # target_type
        if target_type is ...:
            target_type = self.target_type
        else:
            target_type = validate_target_type(target_type)
        
        # target_user
        if target_user is ...:
            target_user = self.target_user
        else:
            target_user = validate_target_user(target_user)
        
        # temporary
        if temporary is ...:
            temporary = self.temporary
        else:
            temporary = validate_temporary(temporary)
        
        # Construct
        new = object.__new__(type(self))
        
        new.approximate_online_count = 0
        new.approximate_user_count = 0
        new.channel = None
        new.code = ''
        new.created_at = None
        new.flags = flags
        new.guild = None
        new.inviter = ZEROUSER
        new.max_age = max_age
        new.max_uses = max_uses
        new.target_application = target_application
        new.target_type = target_type
        new.target_user = target_user
        new.temporary = temporary
        new.type = InviteType.guild
        new.uses = None
        
        return new
        
    
    url = property(module_urls.invite_url)
    
    
    @property
    def id(self):
        """
        Compatibility property with other Discord entities.
        
        Returns
        -------
        id : `int` = `0`
        """
        return 0
    
    
    @property
    def target_application_id(self):
        """
        Returns the invite's target application's identifier.
        
        Returns
        -------
        target_application_id : `int`
        """
        target_application = self.target_application
        if target_application is None:
            target_application_id = 0
        else:
            target_application_id = target_application.id
        
        return target_application_id
    
    
    @property
    def channel_id(self):
        """
        Returns the invite's channel's identifier.
        """
        channel = self.channel
        if channel is None:
            channel_id = 0
        else:
            channel_id = channel.id
        
        return channel_id
    
    
    @property
    def guild_id(self):
        """
        Returns the invite's guild's identifier.
        
        Returns
        -------
        guild_id : `int`
        """
        guild = self.guild
        if guild is None:
            guild_id = 0
        else:
            guild_id = guild.id
        
        return guild_id
    
    
    @property
    def inviter_id(self):
        """
        Returns the invite's creator's identifier.
        
        Returns
        -------
        user_id : `int`
        """
        return self.inviter.id
    
    
    @property
    def target_user_id(self):
        """
        Returns the invite's target user's identifier.
        
        Returns
        -------
        target_user_id : `int`
        """
        target_user = self.target_user
        if target_user is None:
            target_user_id = 0
        else:
            target_user_id = target_user.id
        
        return target_user_id
    
    
    @property
    def partial(self):
        """
        Returns whether the invite is partial.
        
        Since it is not possible to check whether the invite is up to date, it will always return `True` for
        non-template invites.
        
        Returns
        -------
        partial : `bool`
        """
        return False if self.code else True
    
    
    @property
    def stage(self):
        """
        Returns the invite's stage.
        
        Deprecated and will be removed in 2024 February.
        
        Returns
        -------
        stage : `None`, ``InviteStage``
        """
        warn(
            f'`{self.__class__.__name__}.stage` is deprecated and will be removed in 2024 February.',
            FutureWarning,
            stacklevel = 2,
        )
        
        return None
    
    
    @property
    def nsfw_level(self):
        """
        Returns the invite's guild's nsfw level if applicable.
        
        Deprecated and will be removed in 2024 February.
        
        Returns
        -------
        stage : ``NsfwLevel``
        """
        warn(
            f'`{self.__class__.__name__}.nsfw_level` is deprecated and will be removed in 2024 February.',
            FutureWarning,
            stacklevel = 2,
        )
        
        guild = self.guild
        if (guild is None):
            nsfw_level = NsfwLevel.none
        else:
            nsfw_level = guild.nsfw_level
        
        return nsfw_level



__all__ = ()

from scarletio import Compound

from ...application import Application, Entitlement, SKU
from ...application.entitlement.fields import (
    validate_exclude_ended as validate_entitlement_exclude_ended, validate_guild_id as validate_entitlement_guild_id,
    validate_sku_ids as validate_entitlement_sku_ids, validate_user_id as validate_entitlement_user_id
)
from ...application.entitlement.utils import ENTITLEMENT_FIELD_CONVERTERS
from ...http import DiscordApiClient
from ...payload_building import build_create_payload
from ...utils import log_time_converter

from ..request_helpers import get_entitlement_id


def _assert__application_id(application_id):
    """
    Asserts the the client's has `.application` synced by asserting whether it's value is not `0`.
    
    Parameters
    ----------
    application_id : `int`
        The client's application id.
    
    Raises
    ------
    AssertionError
        - If the client's application is not yet synced.
    """
    if application_id == 0:
        raise AssertionError(
            'The client\'s application is not yet synced.'
        )
    
    return True


class ClientCompoundApplicationEndpoints(Compound):
    
    api : DiscordApiClient
    application: Application
    
    
    async def entitlement_create(self, entitlement_template = None, **keyword_parameters):
        """
        Creates a new (test) entitlement.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement_template : `None`, ``Entitlement`` = `None`, Optional
            Entitlement to use as a template
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        owner : `None`, ``ClientUserBase``, ``Guild``, `(int | EntitlementOwnerType, int | str)`, Optional (Keyword only)
            The entitlement's owner. `owner_id` and `owner_type` combined.
        
        owner_id : `int`, Optional (Keyword only)
            The entitlement's owner's identifier.
        
        owner_type : ``EntitlementOwnerType``, `int`, Optional (Keyword only)
            The entitlement's owner's type.
        
        sku : `int`, ``SKU``, Optional (Keyword only)
            Alternative for `sku_id`.
        
        sku_id : `int`, ``SKU``, Optional (Keyword only)
            Stock keeping unit to create the entitlement for.
        
        Returns
        -------
        entitlement : ``Entitlement``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        data = build_create_payload(entitlement_template, ENTITLEMENT_FIELD_CONVERTERS, keyword_parameters)
        entitlement_data = await self.api.entitlement_create(application_id, data)
        return Entitlement.from_data(entitlement_data)
    
    
    async def entitlement_delete(self, entitlement):
        """
        Deletes the given (test) entitlement.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement : ``Entitlement``, `int`
            The entitlement to delete.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        entitlement_id = get_entitlement_id(entitlement)
        await self.api.entitlement_delete(application_id, entitlement_id)
    
    
    async def entitlement_get_all(self, *, exclude_ended = ..., guild_id = ..., sku_ids = ..., user_id = ...):
        """
        Requests all the entitlements of the client's application's entitlements that satisfies the criteria.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exclude_ended : `bool`, Optional (Keyword only)
            Whether ended entitlements should be excluded. Defaults to `False`.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            Guild identifier to filter for.
        
        sku_ids : `int`, `None`, ``SKU``, `iterable` of `int`, `iterable` of ``SKU`, Optional (Keyword only)
            Stock keeping units to filter for.
        
        user_id : `int`, `None`, ``ClientUserBase``, Optional (Keyword only)
            User identifier to filter for.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        query_parameters = {
            'limit': 100,
            'after': 0,
        }
        
        if user_id is not ...:
            user_id = validate_entitlement_user_id(user_id)
            if user_id:
                query_parameters['user_id'] = user_id
        
        if guild_id is not ...:
            guild_id = validate_entitlement_guild_id(guild_id)
            if guild_id:
                query_parameters['guild_id'] = guild_id
        
        if exclude_ended is not ...:
            exclude_ended = validate_entitlement_exclude_ended(guild_id)
            if exclude_ended:
                query_parameters['exclude_ended'] = exclude_ended
        
        if sku_ids is not None:
            sku_ids = validate_entitlement_sku_ids(sku_ids)
            if sku_ids is not None:
                query_parameters['sku_ids'] = ','.join(str(sku_id) for sku_id in sku_ids)
        
        entitlements = []
        
        while True:
            entitlement_datas = await self.api.entitlement_get_chunk(application_id, query_parameters)
            for entitlement_data in entitlement_datas:
                entitlement = Entitlement.from_data(entitlement_data)
                entitlements.append(entitlement)
                
            if len(entitlement_datas) < 100:
                break
            
            query_parameters['after'] = entitlements[-1].id
            continue
        
        return entitlements
    
    
    async def entitlement_get_chunk(
        self,
        *,
        after = ...,
        before = ...,
        exclude_ended = ...,
        guild_id = ...,
        limit = ...,
        sku_ids = ...,
        user_id = ...,
    ):
        """
        Requests a chunk of the client's application's entitlements that satisfies the criteria.
        
        This method is a coroutine.
        
        Parameters
        ----------
        after : `int`, ``DiscordEntity``, `datetime`, Optional (Keyword only)
            The timestamp after the entitlements were created.
        
        before : `int`, ``DiscordEntity``, `datetime`, Optional (Keyword only)
            The timestamp before the entitlements were created.
        
        exclude_ended : `bool`, Optional (Keyword only)
            Whether ended entitlements should be excluded. Defaults to `False`.
        
        guild_id : `int`, `None`, ``Guild``, Optional (Keyword only)
            Guild identifier to filter for.
        
        limit : `int`, Optional (Keyword only)
            Up to how much entitlements should be requested. Can be in range `[1, 100]`. Defaults to `100`.
        
        sku_ids : `int`, `None`, ``SKU``, `iterable` of `int`, `iterable` of ``SKU`, Optional (Keyword only)
            Stock keeping units to filter for.
        
        user_id : `int`, `None`, ``ClientUserBase``, Optional (Keyword only)
            User identifier to filter for.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        if limit is ...:
            limit = 100
        
        elif isinstance(limit, int):
            if limit < 1:
                limit = 1
            elif limit > 100:
                limit = 100
        
        else:
            raise TypeError(
                f'`limit` can be `None`, `int`, got {limit.__class__.__name__}; {limit!r}.'
            )
        
        query_parameters = {
            'limit': limit,
        }
        
        if (after is not ...):
            query_parameters['after'] = log_time_converter(after)
        
        if (before is not ...):
            query_parameters['before'] = log_time_converter(before)
        
        if exclude_ended is not ...:
            exclude_ended = validate_entitlement_exclude_ended(guild_id)
            if exclude_ended:
                query_parameters['exclude_ended'] = exclude_ended
        
        if guild_id is not ...:
            guild_id = validate_entitlement_guild_id(guild_id)
            if guild_id:
                query_parameters['guild_id'] = guild_id
        
        
        if sku_ids is not None:
            sku_ids = validate_entitlement_sku_ids(sku_ids)
            if sku_ids is not None:
                query_parameters['sku_ids'] = ','.join(str(sku_id) for sku_id in sku_ids)
            
        if user_id is not ...:
            user_id = validate_entitlement_user_id(user_id)
            if user_id:
                query_parameters['user_id'] = user_id
        
        entitlement_datas = await self.api.entitlement_get_chunk(application_id, query_parameters)
        return [Entitlement.from_data(entitlement_data) for entitlement_data in entitlement_datas]
    
    
    async def entitlement_consume(self, entitlement):
        """
        Marks the one-time usage entitlement as consumed.
        
        This method is a coroutine.
        
        Parameters
        ----------
        entitlement : ``Entitlement``, `int`
            The entitlement to consume.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        entitlement_id = get_entitlement_id(entitlement)
        await self.api.entitlement_consume(application_id, entitlement_id)
    
    
    async def sku_get_all(self):
        """
        Requests the client's application's stock keeping unit.
        
        This method is a coroutine.
        
        Returns
        -------
        skus : `list` of ``SKU`
            The received stock keeping units.
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        application_id = self.application.id
        assert _assert__application_id(application_id)
        
        sku_datas = await self.api.sku_get_all(application_id)
        return [SKU.from_data(sku_data) for sku_data in sku_datas]



__all__ = ()

from warnings import warn

from scarletio import Compound

from ...application import Application
from ...bases import maybe_snowflake
from ...channel import Channel
from ...exceptions import DiscordException, ERROR_CODES
from ...guild import Guild
from ...http import DiscordApiClient
from ...invite import Invite, InviteTargetType
from ...invite.invite.fields import (
    validate_code, validate_max_age, validate_max_uses, validate_temporary, validate_unique
)
from ...invite.invite.utils import INVITE_GUILD_FIELD_CONVERTERS
from ...payload_building import build_create_payload
from ...permission.permission import PERMISSION_MASK_CREATE_INSTANT_INVITE
from ...user import ClientUserBase

from ..request_helpers import get_channel_id, get_guild_and_id, get_guild_id


def _iter_channels_in_preference_order(guild):
    """
    Iterates over a guild's channel in order of preference.
    
    Parameters
    ----------
    guild : ``Guild``
        The guild to iterate its channels of.
    
    Yields
    ------
    channel : ``Channel``
    """
    channel = guild.rules_channel
    if (channel is not None):
        yield channel 
    
    channel = guild.system_channel
    if (channel is not None):
        yield channel
    
    channel = guild.widget_channel
    if (channel is not None):
        yield channel
    
    for channel in guild.channel_list_flattened:
        if channel.is_in_group_invitable():
            yield channel


def _get_preferred_channel_for_invite(client, guild):
    """
    Gets the preferred channel to create invite for.
    
    Parameters
    ----------
    client : ``Client``
        The client who is getting the channel.
    guild : ``Guild``
        The guild to get the channel of.
    
    Returns
    -------
    channel : `None`, ``Channel``
    """
    for channel in _iter_channels_in_preference_order(guild):
        if channel.cached_permissions_for(client) & PERMISSION_MASK_CREATE_INSTANT_INVITE:
            return channel


class ClientCompoundInviteEndpoints(Compound):
    
    api : DiscordApiClient
    
    
    async def invite_get_vanity(self, guild):
        """
        Returns the vanity invite of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's invite will be returned.
        
        Returns
        -------
        invite : `None`, ``Invite``
            The vanity invite of the `guild`, `None` if it has no vanity invite.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        invite_data_vanity = await self.api.invite_get_vanity(guild_id)
        invite_data = await self.api.invite_get(invite_data_vanity['code'], {'with_counts': True})
        invite_data['uses'] = invite_data_vanity.get('uses', None)
        return Invite.from_data(invite_data)
    
    
    async def vanity_invite_edit(self, guild, vanity_code, *, reason = None):
        """
        Deprecated and will be removed in 2023 February. Please use ``.invite_edit_vanity`` instead.
        """
        warn(
            (
                f'`{self.__class__.__name__}.vanity_invite_edit` is deprecated and will be removed at 2024 February. '
                f'Please use `.invite_edit_vanity` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        return await self.invite_edit_vanity(guild, vanity_code, reason = reason)
    
    
    async def invite_edit_vanity(self, guild, vanity_code, *, reason = None):
        """
        Edits the given guild's vanity invite's code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild, what's invite will be edited.
        vanity_code : `str`
            The new code of the guild's vanity invite.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the guild's audit logs.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `vanity_code` was not given as `str`.
        """
        guild_id = get_guild_id(guild)
        
        vanity_code = validate_code(vanity_code)
        
        await self.api.invite_edit_vanity(guild_id, {'code': vanity_code}, reason)
    
    
    async def invite_create(self, channel, invite_template = None, **keyword_parameters):
        """
        Creates an invite at the given channel with the given parameters.
        
        To create stream invite pass the `target_type` parameter as `InviteTargetTypes.stream` and use
        either the `target_user` or `target_user_id` to define the streamer.
        
        To create an embedded application invite pass the `target_type` parameter as
        `InviteTargetTypes.embedded_application` and use either the `target_application` or `target_application_id` to
        define the application.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel of the created invite.
        
        invite_template : `None`, ``Invite`` = `None`, Optional
            Invite entity to use as a template.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the invite with.
        
        Other Parameters
        ----------------
        max_age : `int`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        
        max_uses : `int`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        
        target_application : `int`, ``Application``, Optional (Keyword only)
            Alternative for `target_application_id`.
        
        target_application_id : `int`, ``Application``, Optional (Keyword only)
            The invite's target application.
        
        target_type : ``InviteTargetType``, `int`, Optional (Keyword only)
            The invite's target type.
        
        target_user : `int`, ``ClientUserBase``, Optional (Keyword only)
            Alternative for `target_user`.
        
        target_user_id : `int`, ``ClientUserBase``, Optional (Keyword only)
            The target of the invite if applicable.
        
        temporary : `bool`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        unique : `bool`, Optional (Keyword only)
            Whether the created invite should be unique.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(channel, Channel.is_in_group_invitable)
        data = build_create_payload(invite_template, INVITE_GUILD_FIELD_CONVERTERS, keyword_parameters)
        invite_data = await self.api.invite_create(channel_id, data)
        return Invite.from_data(invite_data)
    
    
    # 'target_user_id' :
    #     DiscordException BAD REQUEST (400), code = 50035: Invalid Form Body
    #     target_user_type.GUILD_INVITE_INVALID_TARGET_USER_TYPE('Invalid target user type')
    #
    # 'target_type', as 0:
    #     DiscordException BAD REQUEST (400), code = 50035: Invalid Form Body
    #     target_user_type.BASE_TYPE_CHOICES('Value must be one of (1,).')
    #
    # 'target_type', as 1:
    #     DiscordException BAD REQUEST (400), code = 50035: Invalid Form Body
    #     target_user_type.GUILD_INVITE_INVALID_TARGET_USER_TYPE('Invalid target user type')
    #
    # 'target_user_id' and 'target_user_type' together:
    #     DiscordException BAD REQUEST (400), code = 50035: Invalid Form Body
    #     target_user_id.GUILD_INVITE_INVALID_STREAMER('The specified user is currently not streaming in this channel')
    #
    # 'target_user_id' and 'target_user_type' with not correct channel:
    #     DiscordException BAD REQUEST (400), code = 50035: Invalid Form Body
    #     target_user_id.GUILD_INVITE_INVALID_STREAMER('The specified user is currently not streaming in this channel')
    
    async def stream_invite_create(self, guild, user, *, max_age = 0, max_uses = 0, unique = True, temporary = False):
        """
        Creates an STREAM invite at the given guild for the specific user. The user must be streaming at the guild,
        when the invite is created.
        
        Deprecated and will be removed in 2024 February.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``
            The guild where the user streams.
        user : ```ClientUserBase``, `int`
            The streaming user.
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - If `user` was not given neither as ``ClientUserBase`` neither as `int`.
            - If `guild` is not ``Guild``.
        ValueError
            - If the user is not streaming at the guild.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        warn(
            (
                f'`{self.__class__.__name__}.stream_invite_create` is deprecated and will be removed at 2024 February. '
                f'Please use `.invite_create` with `target_type` and `target_user` parameters instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        if not isinstance(guild, Guild):
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        if isinstance(user, ClientUserBase):
            user_id = user.id
        else:
            user_id = maybe_snowflake(user)
            if user_id is None:
                raise TypeError(
                    f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
                )
        
        try:
            voice_state = guild.voice_states[user_id]
        except KeyError:
            raise ValueError(
                f'The user must stream at a voice channel of the guild. Got user = {user!r}; guild = {guild!r}.'
            ) from None
        
        if not voice_state.self_stream:
            raise ValueError(
                f'The user must stream at a voice channel of the guild. Got user = {user!r}; guild = {guild!r}.'
            )
        
        max_age = validate_max_age(max_age)
        max_uses = validate_max_uses(max_uses)
        unique = validate_unique(unique)
        temporary = validate_temporary(temporary)
        
        data = {
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique,
            'target_user_id': user_id,
            'target_type': InviteTargetType.stream.value,
        }
        
        data = await self.api.invite_create(voice_state.channel.id, data)
        return Invite.from_data(data)
    
    # Could not find correct application:
    #    DiscordException Bad Request (400), code = 50035: Invalid Form Body
    #    target_application_id.GUILD_INVITE_INVALID_APPLICATION('The specified application is not embedded')
    
    async def application_invite_create(
        self, channel, application, *, max_age = 0, max_uses = 0, unique = True, temporary = False
    ):
        """
        Creates an EMBEDDED_APPLICATION invite to the specified voice channel. The application must have must have
        `embedded` flag.
        
        Deprecated and will be removed in 2024 February.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int
            The target channel of the invite.
        application : ``Application``, `int`
            The embedded application to open in the voice channel.
            
            > The application must have `EMBEDDED_APPLICATION` flag.
        
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire.
            
            > If given as `0` (so by default) then the created invite will never expire.
        
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used.
            
            > If given as `0` (so by default) then the created invite will have no use limit.
        
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - If `channel` was not given neither as ``Channel``, neither as `int`.
            - If `application` was not given neither as ``Application`` nor as`int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        warn(
            (
                f'`{self.__class__.__name__}.application_invite_create` is deprecated and will be removed at 2024 '
                f'February. '
                f'Please use `.invite_create` with `target_type` and `target_user` parameters instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        channel_id = get_channel_id(channel, Channel.is_guild_voice)
        
        if isinstance(application, Application):
            application_id = application.id
        else:
            application_id = maybe_snowflake(application)
            if application_id is None:
                raise TypeError(
                    f'`application` can be `{Application.__name__}`, `int`, got '
                    f'{application.__class__.__name__}; {application!r}.'
                )
        
        max_age = validate_max_age(max_age)
        max_uses = validate_max_uses(max_uses)
        unique = validate_unique(unique)
        temporary = validate_temporary(temporary)
        
        data = {
            'max_age': max_age,
            'max_uses': max_uses,
            'temporary': temporary,
            'unique': unique,
            'target_application_id': application_id,
            'target_type': InviteTargetType.embedded_application.value,
        }
        
        data = await self.api.invite_create(channel_id, data)
        return Invite.from_data(data)
    
    
    async def invite_create_preferred(self, guild, **keyword_parameters):
        """
        Creates an invite to the guild's preferred channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild . ``Guild``
            The guild the invite will be created to.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to describe the created invite.
        
        Other Parameters
        ----------------
        max_age : `int` = `0`, Optional (Keyword only)
            After how much time (in seconds) will the invite expire. Defaults is never.
        
        max_uses : `int` = `0`, Optional (Keyword only)
            How much times can the invite be used. Defaults to unlimited.
        
        unique : `bool` = `True`, Optional (Keyword only)
            Whether the created invite should be unique.
        
        temporary : `bool` = `False`, Optional (Keyword only)
            Whether the invite should give only temporary membership.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            - if `guild` is not ``Guild`˛
        ValueError
            If the guild has no channel to create invite to.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if not isinstance(guild, Guild):
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        channel = _get_preferred_channel_for_invite(self, guild)
        if channel is None:
            raise ValueError(
                f'The guild has no available channels to create invite for (or lack of permissions).'
                f'Got guild = {guild!r}.'
            )
        
        try:
            return (await self.invite_create(channel, **keyword_parameters))
        except DiscordException as err:
            if err.code in (
                ERROR_CODES.unknown_channel, # the channel was deleted meanwhile
                ERROR_CODES.missing_permissions, # permissions changed meanwhile
                ERROR_CODES.missing_access, # client removed
            ):
                return None
            raise
    
    
    async def invite_get(self, invite):
        """
        Requests a partial invite with the given code.
        
        This method is a coroutine.
        
        Parameters
        ----------
        invite : ``Invite``, `str`
            The invites code.
        
        Returns
        -------
        invite : ``Invite``
        
        Raises
        ------
        TypeError
            If `invite` was not given neither ``Invite`` nor `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            If `invite_code` was not given as `str`.
        """
        if isinstance(invite, Invite):
            invite_code = invite.code
        
        elif isinstance(invite, str):
            invite_code = invite
            invite = None
        
        else:
            raise TypeError(
                f'`invite`` can be `{Invite.__name__}`, `str`, got {invite.__class__.__name__}; '
                f'{invite!r}.'
            )
        
        invite_data = await self.api.invite_get(invite_code, {'with_counts': True})
        
        if invite is None:
            invite = Invite.from_data(invite_data)
        else:
            invite._update_attributes(invite_data)
        
        return invite
    
    
    async def invite_get_all_guild(self, guild):
        """
        Gets the invites of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's invites will be requested.
        
        Returns
        -------
        invites : `list` of ``Invite`` objects
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        
        invite_datas = await self.api.invite_get_all_guild(guild_id)
        return [Invite.from_data(invite_data) for invite_data in invite_datas]
    
    
    async def invite_get_all_channel(self, channel):
        """
        Gets the invites of the given channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel, what's invites will be requested.
        
        Returns
        -------
        invites : `list` of ``Invite`` objects
        
        Raises
        ------
        TypeError
            If `channel` was not given neither as ``Channel``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        while True:
            if isinstance(channel, Channel):
                if channel.is_in_group_invitable() or channel.partial:
                    channel_id = channel.id
                    break
            
            else:
                channel_id = maybe_snowflake(channel)
                if (channel_id is not None):
                    break
            
            raise TypeError(
                f'`channel` can be an invitable ``Channel``, `int`, got {channel.__class__.__name__}; {channel!r}.'
            )
        
        invite_datas = await self.api.invite_get_all_channel(channel_id)
        return [Invite.from_data(invite_data) for invite_data in invite_datas]
    
    
    async def invite_delete(self, invite, *, reason = None):
        """
        Deletes the given invite.
        
        This method is a coroutine.
        
        Parameters
        ----------
        invite : ``Invite``
            The invite to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Shows up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `invite` was not given neither ``Invite`` nor `str`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if isinstance(invite, Invite):
            invite_code = invite.code
        
        elif isinstance(invite, str):
            invite_code = invite
            invite = None
        
        else:
            raise TypeError(
                f'`invite`` can be `{Invite.__name__}`, `str`, got {invite.__class__.__name__}; {invite!r}.'
            )
        
        invite_data = await self.api.invite_delete(invite_code, reason)
        
        if invite is None:
            invite = Invite.from_data(invite_data)
        else:
            invite._update_attributes(invite_data)
        
        return invite



__all__ = ('ReactionType',)

from ...bases import Preinstance as P, PreinstancedBase


class ReactionType(PreinstancedBase):
    """
    Represents a reaction's type.
    
    Attributes
    ----------
    name : `str`
        The default name of the reaction type.
    value : `int`
        The discord side identifier value of the reaction type.    Class Attributes
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ReactionType``) items
        Stores the predefined ``ReactionType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The reaction type' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the reaction type.
    
    Every predefined reaction type can be accessed as class attribute as well:
    
    +-----------------------+---------------+-------+
    | Class attribute name  | name          | value |
    +======================+================+=======+
    | standard              | standard      | 0     |
    +-----------------------+---------------+-------+
    | burst                 | status        | 1     |
    +-----------------------+---------------+-------+
    """
    __slots__ = ()
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    standard = P(0, 'standard')
    burst = P(1, 'burst')



__all__ = ('EntitySelectDefaultValueType',)

from ...bases import Preinstance as P, PreinstancedBase


class EntitySelectDefaultValueType(PreinstancedBase):
    """
    Represents an entity select default option's type.
    
    Attributes
    ----------
    name : `str`
        The name of the entity select default option type.
    value : `str`
        The Discord side identifier value of the entity select default option type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``EntitySelectDefaultValueType``) items
        Stores the created entity select default option type instances. This container is accessed when translating a
        Discord entity select default option type's value to it's representation.
    VALUE_TYPE : `type` = `str`
        The entity select default option types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the entity select default option types.
    
    Every predefined entity select default option type can be accessed as class attribute as well:
    +---------------------------+---------------------------+---------------+
    | Class attribute name      | name                      | value         |
    +===========================+===========================+===============+
    | none                      | none                      | `''`          |
    +---------------------------+---------------------------+---------------+
    | channel                   | channel                   | `'channel'`   |
    +---------------------------+---------------------------+---------------+
    | role                      | role                      | `'role'`      |
    +---------------------------+---------------------------+---------------+
    | user                      | user                      | `'user'`      |
    +---------------------------+---------------------------+---------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ()
    
    # predefined
    none = P('', 'none')
    channel = P('channel', 'channel')
    role = P('role', 'role')
    user = P('user', 'user')



__all__ = (
    'ApplicationDiscoverabilityState', 'ApplicationExplicitContentFilterLevel', 'ApplicationIntegrationType',
    'ApplicationInteractionEventType', 'ApplicationInteractionVersion', 'ApplicationInternalGuildRestriction',
    'ApplicationMonetizationState', 'ApplicationRPCState', 'ApplicationStoreState', 'ApplicationType',
    'ApplicationVerificationState' 
)

from ...bases import Preinstance as P, PreinstancedBase


class ApplicationDiscoverabilityState(PreinstancedBase):
    """
    Represents an application's discoverability state.
    
    Attributes
    ----------
    name : `str`
        The name of the application discoverability state.
    value : `int`
        The Discord side identifier value of the application discoverability state.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationDiscoverabilityState``) items
        Stores the created application discoverability state instances. This container is accessed when translating a
        Discord application discoverability state's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application discoverability states' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application discoverability states.
    
    Every predefined application discoverability state can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | ineligible                | ineligible                | 1     |
    +---------------------------+---------------------------+-------+
    | not_discoverable          | not discoverable          | 2     |
    +---------------------------+---------------------------+-------+
    | discoverable              | discoverable              | 3     |
    +---------------------------+---------------------------+-------+
    | featurable                | featurable                | 4     |
    +---------------------------+---------------------------+-------+
    | blocked                   | blocked                   | 5     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    ineligible = P(1, 'ineligible')
    not_discoverable = P(2, 'not discoverable')
    discoverable = P(3, 'discoverable')
    featurable = P(4, 'featurable')
    blocked = P(5, 'blocked')


class ApplicationExplicitContentFilterLevel(PreinstancedBase):
    """
    Represents an application's explicit content filter level.
    
    Attributes
    ----------
    name : `str`
        The name of the application explicit content filter level.
    value : `int`
        The Discord side identifier value of the application explicit content filter level.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationExplicitContentFilterLevel``) items
        Stores the created application explicit content filter level instances. This container is accessed when translating a
        Discord application explicit content filter level's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application explicit content filter levels' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application explicit content filter levels.
    
    Every predefined application explicit content filter level can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | filtered                  | filtered                  | 1     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    filtered = P(1, 'filtered')


class ApplicationIntegrationType(PreinstancedBase):
    """
    Represents how an application can be integrated.
    
    Attributes
    ----------
    name : `str`
        The name of the application integration type.
    value : `int`
        The Discord side identifier value of the application integration type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationIntegrationType``) items
        Stores the created application integration type instances. This container is accessed when translating a
        Discord application integration type's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application integration types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application integration types.
    
    Every predefined application integration type can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | guild_install             | guild install             | 0     |
    +---------------------------+---------------------------+-------+
    | user_install              | user install              | 1     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    guild_install = P(0, 'guild install')
    user_install = P(1, 'user install')


class ApplicationInteractionEventType(PreinstancedBase):
    """
    Represents an application's interaction event type.
    
    Attributes
    ----------
    name : `str`
        The name of the application interaction event type.
    value : `int`
        The Discord side identifier value of the application interaction event type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationInteractionEventType``) items
        Stores the created application interaction event type instances. This container is accessed when translating a
        Discord application interaction event type's value to it's representation.
    VALUE_TYPE : `type` = `str`
        The application interaction event types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application interaction event types.
    
    Every predefined application interaction event type can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | `''`  |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ()
    
    # predefined
    none = P('', 'none')


class ApplicationInteractionVersion(PreinstancedBase):
    """
    Represents an application interaction version.
    
    Attributes
    ----------
    name : `str`
        The name of the application interaction version.
    value : `int`
        The Discord side identifier value of the application interaction version.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationInteractionVersion``) items
        Stores the created application interaction version instances. This container is accessed when translating a
        Discord application interaction version's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application interaction versions' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application interaction versions.
    
    Every predefined application interaction version can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | every                     | every                     | 1     |
    +---------------------------+---------------------------+-------+
    | selective                 | selective                 | 2     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    every = P(1, 'every')
    selective = P(2, 'selective')


class ApplicationInternalGuildRestriction(PreinstancedBase):
    """
    Represents an application's internal guild restriction.
    
    > Note that could not find anything about this.
    
    Attributes
    ----------
    name : `str`
        The name of the application internal guild restriction.
    value : `int`
        The Discord side identifier value of the application internal guild restriction.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationInternalGuildRestriction``) items
        Stores the created application internal guild restriction instances. This container is accessed when
        translating a Discord application internal guild restriction's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application internal guild restrictions' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application internal guild restrictions.
    
    Every predefined application internal guild restriction can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | restricted                | restricted                | 1     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    restricted = P(1, 'restricted')


class ApplicationMonetizationState(PreinstancedBase):
    """
    Represents an application's monetization state.
    
    Attributes
    ----------
    value : `int`
        The Discord side identifier value of the monetization state.
    name : `str`
        The default name of the monetization state.
    enabled : `bool`
        Whether monetization is enabled.
    locked : `bool`
        Whether monetization actions are locked.
    settable : `bool`
        Whether the monetization state can be set.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`str`, ``ApplicationMonetizationState``) items
        Stores the predefined monetization states. This container is accessed when translating a Discord side
        identifier of a monetization state. The identifier value is used as a key to get it's wrapper side
        representation.
    VALUE_TYPE : `state` = `str`
        The monetization states' values' state.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the monetization states.
    
    Every predefined monetization state is also stored as a class attribute:
    
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | Class attribute name  | Name                  | Value     | Enabled   | Locked    | Settable  |
    +=======================+=======================+===========+===========+===========+===========+
    | none                  | none                  | 0         | `False`   | `False`   | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | disabled              | disabled              | 1         | `False`   | `False`   | `True`    |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | provisional           | provisional           | 2         | `True`    | `False`   | `True`    |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | rejected              | rejected              | 3         | `False`   | `True`    | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | approved              | approved              | 4         | `True`    | `False`   | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    | blocked               | blocked               | 5         | `False`   | `True`    | `False`   |
    +-----------------------+-----------------------+-----------+-----------+-----------+-----------+
    """
    __slots__ = ('enabled', 'locked', 'settable')
    
    INSTANCES = {}
    VALUE_TYPE = int
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new monetization state with the given value.
        
        Parameters
        ----------
        value : `int`
            The monetization state's identifier value.
        
        Returns
        -------
        self : `instance<cls>`
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.enabled = False
        self.locked = False
        self.settable = False
        
        return self
    
    
    def __init__(self, value, name, enabled, locked, settable):
        """
        Creates a monetization state and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the monetization state.
        name : `str`
            The default name of the monetization state.
        enabled : `bool`
            Whether monetization is enabled.
        locked : `bool`
            Whether monetization actions are locked.
        settable : `bool`
            Whether the monetization state can be set.
        """
        self.value = value
        self.name = name
        self.enabled = enabled
        self.locked = locked
        self.settable = settable
        
        self.INSTANCES[value] = self
    
    # predefined
    none = P(0, 'none', False, False, False)
    disabled = P(1, 'disabled', False, False, True)
    provisional = P(2, 'provisional', True, False, True)
    rejected = P(3, 'rejected', False, True, False)
    approved = P(4, 'approved', True, False, False)
    blocked = P(5, 'blocked', False, True, False)


class ApplicationRPCState(PreinstancedBase):
    """
    Represents an application rpc state.
    
    Attributes
    ----------
    name : `str`
        The name of the application rpc state.
    value : `int`
        The Discord side identifier value of the application rpc state.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationRPCState``) items
        RPCs the created application rpc state instances. This container is accessed when translating a
        Discord application rpc state's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application rpc states' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application rpc states.
    
    Every predefined application rpc state can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | not_submitted             | not submitted             | 1     |
    +---------------------------+---------------------------+-------+
    | submitted                 | submitted                 | 2     |
    +---------------------------+---------------------------+-------+
    | approved                  | approved                  | 3     |
    +---------------------------+---------------------------+-------+
    | rejected                  | rejected                  | 4     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    not_submitted = P(1, 'not submitted')
    submitted = P(2, 'submitted')
    approved = P(3, 'approved')
    rejected = P(4, 'rejected')


class ApplicationStoreState(PreinstancedBase):
    """
    Represents an application store state.
    
    Attributes
    ----------
    name : `str`
        The name of the application store state.
    value : `int`
        The Discord side identifier value of the application store state.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationStoreState``) items
        Stores the created application store state instances. This container is accessed when translating a
        Discord application store state's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application store states' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application store states.
    
    Every predefined application store state can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | no_license                | no license                | 1     |
    +---------------------------+---------------------------+-------+
    | license                   | license                   | 2     |
    +---------------------------+---------------------------+-------+
    | submitted                 | submitted                 | 3     |
    +---------------------------+---------------------------+-------+
    | approved                  | approved                  | 4     |
    +---------------------------+---------------------------+-------+
    | rejected                  | rejected                  | 5     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    no_license = P(1, 'no license')
    license = P(2, 'license')
    submitted = P(3, 'submitted')
    approved = P(4, 'approved')
    rejected = P(5, 'rejected')


class ApplicationType(PreinstancedBase):
    """
    Represents an application type.
    
    Attributes
    ----------
    name : `str`
        The name of the application type.
    value : `int`
        The Discord side identifier value of the application type.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationType``) items
        Stores the created application type instances. This container is accessed when translating a Discord
        application type's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application types' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application types.
    
    Every predefined application type can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | game                      | game                      | 1     |
    +---------------------------+---------------------------+-------+
    | music                     | music                     | 2     |
    +---------------------------+---------------------------+-------+
    | ticketed_event            | ticketed event            | 3     |
    +---------------------------+---------------------------+-------+
    | guild_role_subscription   | guild role subscription   | 4     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    game = P(1, 'game')
    music = P(2, 'music')
    ticketed_event = P(3, 'ticketed event')
    guild_role_subscription = P(4, 'guild role subscription')


class ApplicationVerificationState(PreinstancedBase):
    """
    Represents an application's verification state.
    
    Attributes
    ----------
    name : `str`
        The name of the application verification state.
    value : `int`
        The Discord side identifier value of the application verification state.
        
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationVerificationState``) items
        Stores the created application verification state instances. This container is accessed when translating a
        Discord application verification state's value to it's representation.
    VALUE_TYPE : `type` = `int`
        The application verification states' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the application verification states.
    
    Every predefined application verification state can be accessed as class attribute as well:
    
    +---------------------------+---------------------------+-------+
    | Class attribute name      | name                      | value |
    +===========================+===========================+=======+
    | none                      | none                      | 0     |
    +---------------------------+---------------------------+-------+
    | ineligible                | ineligible                | 1     |
    +---------------------------+---------------------------+-------+
    | not_submitted             | not submitted             | 2     |
    +---------------------------+---------------------------+-------+
    | submitted                 | submitted                 | 3     |
    +---------------------------+---------------------------+-------+
    | approved                  | approved                  | 4     |
    +---------------------------+---------------------------+-------+
    | blocked                   | blocked                   | 5     |
    +---------------------------+---------------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    
    __slots__ = ()
    
    # predefined
    none = P(0, 'none')
    ineligible = P(1, 'ineligible')
    not_submitted = P(2, 'not submitted')
    submitted = P(3, 'submitted')
    approved = P(4, 'approved')
    blocked = P(5, 'blocked')



__all__ = ()

from ...field_parsers import preinstanced_parser_factory
from ...field_putters import preinstanced_putter_factory
from ...field_validators import preinstanced_validator_factory

from .preinstanced import ActivityType

# type

parse_type = preinstanced_parser_factory('type', ActivityType, ActivityType.game)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('type', ActivityType)



__all__ = ()

from functools import partial as partial_func

from .fields import (
    put_allow_into, put_deny_into, put_target_id_into, put_target_into, put_target_type_into, validate_allow,
    validate_deny, validate_target, validate_target_id, validate_target_type
)


PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS = {
    'allow': (validate_allow, put_allow_into),
    'deny': (validate_deny, put_deny_into),
}


PERMISSION_OVERWRITE_FIELD_CONVERTERS = {
    **PERMISSION_OVERWRITE_PERMISSION_FIELD_CONVERTERS,
    'target': (validate_target, partial_func(put_target_into, include_internals = True)),
    'target_id': (validate_target_id, put_target_id_into),
    'target_type': (validate_target_type, put_target_type_into),
}



__all__ = ('LabelType', 'ReleasePhase',)

from ...bases import Preinstance as P, PreinstancedBase


class LabelType(PreinstancedBase):
    """
    Represents a embedded activity's label's type.
    
    Attributes
    ----------
    name : `str`
        The name of the label's type.
    value : `int`
        The identifier value the label's type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``LabelType``) items
        Stores the predefined label's types. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The label's types' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the label's types.
    
    Every predefined label's type can be accessed as class attribute as well:
    
    +-----------------------+-----------+-------+
    | Class attribute name  | Name      | Value |
    +=======================+===========+=======+
    | none                  | none      | 0     |
    +-----------------------+-----------+-------+
    | new                   | new       | 1     |
    +-----------------------+-----------+-------+
    | updated               | updated   | 2     |
    +-----------------------+-----------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ()
    
    none = P(0, 'none')
    new = P(1, 'new')
    updated = P(2, 'updated')


class ReleasePhase(PreinstancedBase):
    """
    Represents the release phase of an embedded activity on a platform.
    
    Attributes
    ----------
    value : `str`
        The Discord side identifier value of the platform.
    name : `str`
        The name of the release phase.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ReleasePhase``) items
        Stores the predefined release phases. This container is accessed when converting Discord side release phase's
        value to it's wrapper side representation.
    VALUE_TYPE : `type` = `str`
        The release phases' values' type.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name of the release phases.
    
    Every predefined release phase can also be accessed as class attribute:
    
    +-----------------------+---------------+---------------+
    | Class attribute name  | name          | value         |
    +=======================+===============+===============+
    | global_launch         | global launch | global_launch |
    +-----------------------+---------------+---------------+
    """
    INSTANCES = {}
    VALUE_TYPE = str
    
    __slots__ = ()

    # predefined
    global_launch = P('global_launch', 'global launch')


ReleasePhase.INSTANCES[''] = ReleasePhase.global_launch



__all__ = ('MessageApplication',)

from ...bases import DiscordEntity, ICON_TYPE_NONE, IconSlot
from ...http import urls as module_urls
from ...precreate_helpers import process_precreate_parameters_and_raise_extra

from .fields import (
    parse_description, parse_id, parse_name, put_description_into, put_id_into, put_name_into,
    validate_description, validate_id, validate_name
)

APPLICATION_COVER = IconSlot(
    'cover',
    'cover_image',
    module_urls.application_cover_url,
    module_urls.application_cover_url_as,
    add_updater = False,
)

APPLICATION_ICON = IconSlot(
    'icon',
    'icon',
    module_urls.application_icon_url,
    module_urls.application_icon_url_as,
    add_updater = False,
)

PRECREATE_FIELDS = {
    'cover': ('cover', APPLICATION_COVER.validate_icon),
    'description': ('description', validate_description),
    'icon': ('icon', APPLICATION_ICON.validate_icon),
    'name': ('name', validate_name),
}


class MessageApplication(DiscordEntity):
    """
    Might be sent with a ``Message`` if it has rich presence-related chat embeds.
    
    Attributes
    ----------
    cover_hash : `int`
        The respective application's store cover image's hash in `uint128`. If the application is sold at Discord,
        this image will be used at the store.
    cover_type : ``IconType``
        The respective application's store cover image's type.
    description : `None` `str`
        The description of the application. Defaults to empty string.
    icon_hash : `int`
        The application's icon's hash as `uint128`.
    icon_type : `IconType`
        The application's icon's type.
    id : `int`
        The application's id.
    name : `str`
        The name of the application. Defaults to empty string.
    """
    __slots__ = ('description', 'name', )
    
    cover = APPLICATION_COVER
    icon = APPLICATION_ICON
    
    def __new__(cls, *, cover = ..., description = ..., icon = ..., name = ...):
        """
        Creates a new message application from the given keyword parameters.
        
        Parameters
        ----------
        cover : `None`, `bytes`, `bytearray`, `memoryview`, ``Icon``, Optional (Keyword only)
            The cover of the message application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        icon : `None`, `bytes`, `bytearray`, `memoryview`, ``Icon``, Optional (Keyword only)
            The icon of the message application.
        
        name : `None, `str`, Optional (Keyword only)
            The name of the application.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # cover
        if cover is ...:
            cover = None
        else:
            cover = cls.cover.validate_icon(cover, allow_data = True)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # icon
        if icon is ...:
            icon = None
        else:
            icon = cls.icon.validate_icon(icon, allow_data = True)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # Construct
        self = object.__new__(cls)
        self.cover = cover
        self.description = description
        self.icon = icon
        self.id = 0
        self.name = name
        return self
    
    
    @classmethod
    def precreate(cls, message_application_id, **keyword_parameters):
        """
        Creates a new message application with the given predefined fields.
        
        > Since message applications are not globally cached, this method is only used for testing.
        
        Parameters
        ----------
        message_application_id : `int`
            The message application's id.
        
        **keyword_parameters : Keyword parameters
            The attributes to set.
        
        Other Parameters
        ----------------
        cover : `None`, ``Icon``, Optional (Keyword only)
            The cover of the message application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        icon : `None`, ``Icon``, Optional (Keyword only)
            The icon of the message application.
        
        name : `None, `str`, Optional (Keyword only)
            The name of the application.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        message_application_id = validate_id(message_application_id)
        
        if keyword_parameters:
            processed = process_precreate_parameters_and_raise_extra(keyword_parameters, PRECREATE_FIELDS)
        else:
            processed = None
        
        self = cls._create_empty(message_application_id)
        
        if (processed is not None):
            for item in processed:
                setattr(self, *item)
        
        return self
    
    
    @classmethod
    def _create_empty(cls, message_application_id):
        """
        Creates a new message application with it's defaults attributes set.
        
        Parameters
        ----------
        message_application_id : `int`
            The message application's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.cover_type = ICON_TYPE_NONE
        self.cover_hash = 0
        self.description = None
        self.icon_type = ICON_TYPE_NONE
        self.icon_hash = 0
        self.id = message_application_id
        self.name = ''
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new message application instance with the given application data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Application data included within message payload.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._set_cover(data)
        self.description = parse_description(data)
        self._set_icon(data)
        self.id = parse_id(data)
        self.name = parse_name(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the message application to json serializable data.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        type(self).cover.put_into(self.cover, data, defaults, as_data = not include_internals)
        put_description_into(self.description, data, defaults)
        type(self).icon.put_into(self.icon, data, defaults, as_data = not include_internals)
        put_name_into(self.name, data, defaults)
        
        if include_internals:
            put_id_into(self.id, data, defaults)
        
        return data
    
    
    @property
    def partial(self):
        """
        Returns whether the message application is partial.
        
        Returns
        -------
        partial : `bool`
        """
        return (self.id == 0)
    
    
    def __eq__(self, other):
        """Returns whether the two message applications are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def __ne__(self, other):
        """Returns whether the two message applications not equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return not self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `instance<type<<self>>`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        # id
        self_id = self.id
        other_id = other.id
        if (self_id and other_id) and (self_id != other_id):
            return False
        
        # cover_hash
        if self.cover_hash != other.cover_hash:
            return False
        
        # cover_type
        if self.cover_type != other.cover_type:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # icon_hash
        if self.icon_hash != other.icon_hash:
            return False
        
        # icon_type
        if self.icon_type != other.icon_type:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the message application's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        message_application_id = self.id
        if message_application_id:
            repr_parts.append(' id = ')
            repr_parts.append(repr(message_application_id))
        else:
            repr_parts.append(' (partial)')
        
        name = self.name
        if name:
            repr_parts.append(', name = ')
            repr_parts.append(repr(name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the message application's hash value."""
        # These entities are not cached, so we wont use their `id` if applicable.
        hash_value = 0
        
        # cover
        hash_value ^= hash(self.cover)
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # icon
        hash_value ^= hash(self.icon)
        
        # id
        hash_value ^= self.id
        
        # name
        name = self.name
        if (description is None) or (description != name):
            hash_value ^= hash(name)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the message application returning a new partial one.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.cover_hash = self.cover_hash
        new.cover_type = self.cover_type
        new.description = self.description
        new.icon_hash = self.icon_hash
        new.icon_type = self.icon_type
        new.id = 0
        new.name = self.name
        return new


    def copy_with(self, *, cover = ..., description = ..., icon = ..., name = ...):
        """
        Copies the message application with the given fields returning a new partial one.
        
        Parameters
        ----------
        cover : `None`, `bytes`, `bytearray`, `memoryview`, ``Icon``, Optional (Keyword only)
            The cover of the message application.
        
        description : `None`, `str`, Optional (Keyword only)
            The description of the application.
        
        icon : `None`, `bytes`, `bytearray`, `memoryview`, ``Icon``, Optional (Keyword only)
            The icon of the message application.
        
        name : `None, `str`, Optional (Keyword only)
            The name of the application.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # cover
        if cover is ...:
            cover = self.cover
        else:
            cover = type(self).cover.validate_icon(cover, allow_data = True)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # icon
        if icon is ...:
            icon = self.icon
        else:
            icon = type(self).icon.validate_icon(icon, allow_data = True)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # Construct
        new = object.__new__(type(self))
        new.cover = cover
        new.description = description
        new.icon = icon
        new.id = 0
        new.name = name
        return new



__all__ = ()

from ..application import Entitlement
from ..application_command import ApplicationCommand
from ..auto_moderation import AutoModerationRule
from ..bases import maybe_snowflake, maybe_snowflake_pair, maybe_snowflake_token_pair
from ..channel import Channel, ForumTag, PermissionOverwrite
from ..core import (
    APPLICATION_COMMANDS, AUTO_MODERATION_RULES, CHANNELS, FORUM_TAGS, GUILDS, MESSAGES, ROLES, SCHEDULED_EVENTS,
    SOUNDBOARD_SOUNDS, STICKERS, STICKER_PACKS, USERS
)
from ..emoji import Emoji, Reaction, ReactionType
from ..guild import Guild
from ..message import Message
from ..oauth2 import Achievement, Oauth2Access, Oauth2User
from ..role import Role
from ..poll import PollAnswer
from ..scheduled_event import ScheduledEvent
from ..soundboard import SoundboardSound
from ..stage import Stage
from ..sticker import Sticker, StickerPack
from ..user import ClientUserBase
from ..webhook import Webhook


def validate_message_to_delete(message):
    """
    Validates a message to delete.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message to validate for deletion.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier where the message is.
    message_id : `int`
        The message's identifier.
    message : `None`, ``Message``
        The referenced message if found.
    
    Raises
    ------
    TypeError
        If message was not given neither as ``Message``, `tuple` (`int`, `int`).
    """
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
        
        message = MESSAGES.get(message, None)
    
    return channel_id, message_id, message


def get_channel_id(channel, type_checker):
    """
    Gets the channel's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `int`
        The channel, or it's identifier.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, Channel):
        if type_checker(channel) or channel.partial:
            return channel.id
    
    else:
        channel_id = maybe_snowflake(channel)
        if (channel_id is not None):
            return channel_id
        
    raise TypeError(
        f'`channel` can be `{Channel.__name__}`, `int`,  passing the `{type_checker.__name__}` check, '
        f'got {channel.__class__.__name__}; {channel!r}.'
    )


def get_channel_guild_id_and_id(channel, type_checker):
    """
    Gets the channel's and it's guild's identifier from the given channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `tuple` (`int`, `int`)
        The role, or `guild-id`, `role-id` pair.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    snowflake_pair : `tuple` (`int`, `int`)
        The channel's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    if isinstance(channel, Channel):
        if type_checker(channel) or channel.partial:
            return channel.guild_id, channel.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(channel)
        if (snowflake_pair is not None):
            return snowflake_pair
    
    raise TypeError(
        f'`channel` can be `{Channel.__name__}`, `tuple` (`int`, `int`),  passing the `{type_checker.__name__}` check, '
        f'got {channel.__class__.__name__}; {channel!r}.'
    )


def get_channel_and_id(channel, type_checker):
    """
    Gets thread channel and it's identifier from the given thread channel or of it's identifier.
    
    Parameters
    ----------
    channel : ``Channel``, `int`
        The channel, or it's identifier.
    type_checker : `FunctionType`
        Type checker for `channel`.
    
    Returns
    -------
    channel : `None`, ``Channel``
        The channel if found.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `channel`'s type is incorrect.
    """
    while True:
        if isinstance(channel, Channel):
            if type_checker(channel) or channel.partial:
                channel_id = channel.id
                break
        
        else:
            channel_id = maybe_snowflake(channel)
            if (channel_id is not None):
                try:
                    channel = CHANNELS[channel_id]
                except KeyError:
                    channel = None
                    break
                
                if type_checker(channel) or channel.partial:
                    channel_id = channel.id
                    break
        
        raise TypeError(
            f'`channel` can be `{Channel.__name__}`, `int`,  passing the `{type_checker.__name__}` check, '
            f'got {channel.__class__.__name__}; {channel!r}.'
        )
    
    return channel, channel_id


def get_stage_and_channel_id(stage):
    """
    Gets the stage an its channel's identifier.
    
    Parameters
    ----------
    stage : `Stage`, ``Channel``, `int`
        The stage, or it's identifier.
    
    Returns
    -------
    stage : `None`, ``Stage``
        The identified stage if any.
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `stage`'s type is incorrect.
    """
    while True:
        if isinstance(stage, Stage):
            channel_id = stage.channel.id
            break
        
        elif isinstance(stage, Channel):
            if stage.is_guild_stage() or stage.partial:
                channel_id = stage.id
                stage = None
                break
        
        else:
            channel_id = maybe_snowflake(stage)
            if (channel_id is not None):
                stage = None
                break
        
        raise TypeError(
            f'`stage` can be `{Stage.__name__}`, `{Channel.__name__}`, `int`, got '
            f'{stage.__class__.__name__}; {stage!r}.'
        )
    
    return stage, channel_id


def get_stage_channel_id(stage):
    """
    Gets stage's channel's identifier from the given stage or of it's identifier.
    
    Parameters
    ----------
    stage : `Stage`, ``Channel``, `int`
        The stage, or it's identifier.
    
    Returns
    -------
    channel_id : `int`
        The channel's identifier.
    
    Raises
    ------
    TypeError
        If `stage`'s type is incorrect.
    """
    while True:
        if isinstance(stage, Stage):
            channel_id = stage.channel.id
            break
        
        elif isinstance(stage, Channel):
            if stage.is_guild_stage() or stage.partial:
                channel_id = stage.id
                break
        
        else:
            channel_id = maybe_snowflake(stage)
            if (channel_id is not None):
                break
        
        raise TypeError(
            f'`stage` can be `{Stage.__name__}`, `{Channel.__name__}`, `int`, got '
            f'{stage.__class__.__name__}; {stage!r}.'
        )
    
    return channel_id


def get_user_id(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if isinstance(user, ClientUserBase):
        return user.id
    
    user_id = maybe_snowflake(user)
    if (user_id is not None):
        return user_id
    
    raise TypeError(
        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
    )


def get_user_and_id(user):
    """
    Gets user and it's identifier from the given user or of it's identifier.
    
    Parameters
    ----------
    user : ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user : `None`, ``ClientUserBase``
        The user if found.
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    while True:
        if isinstance(user, ClientUserBase):
            user_id = user.id
            break
        
        user_id = maybe_snowflake(user)
        if (user_id is not None):
            try:
                user = USERS[user_id]
            except KeyError:
                user = None
                break
            
            if isinstance(user, ClientUserBase):
                break
        
        raise TypeError(
            f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
        )
        
    return user, user_id


def get_user_id_nullable(user):
    """
    Gets user identifier from the given user or of it's identifier.
    
    > The user can be `None`. At that case `user_id` will default to `0`.
    
    Parameters
    ----------
    user : `None`, ``ClientUserBase``, `int`
        The user, or it's identifier.
    
    Returns
    -------
    user_id : `int`
        The user's identifier.
    
    Raises
    ------
    TypeError
        If `user`'s type is incorrect.
    """
    if user is None:
        return 0
    
    if isinstance(user, ClientUserBase):
        return user.id
    
    user_id = maybe_snowflake(user)
    if (user_id is not None):
        return user_id
        
    raise TypeError(
        f'`user` can be `{ClientUserBase.__name__}`, `int`, got {user.__class__.__name__}; {user!r}.'
    )


def get_guild_id(guild):
    """
    Gets the guild's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        return guild.id
    
    guild_id = maybe_snowflake(guild)
    if (guild_id is not None):
        return guild_id
        
    raise TypeError(
        f'`guild` can be `{Guild.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
    )


def get_guild_and_id(guild):
    """
    Gets the guild and it's identifier from the given guild or of it's identifier.
    
    Parameters
    ----------
    guild : ``Guild``, `int`
        The guild, or it's identifier.
    
    Returns
    -------
    guild : `None`, ``Guild``
        The guild if found.
    guild_id : `int`
        The guild's identifier.
    
    Raises
    ------
    TypeError
        If `guild`'s type is incorrect.
    """
    if isinstance(guild, Guild):
        guild_id = guild.id
    else:
        guild_id = maybe_snowflake(guild)
        if guild_id is None:
            raise TypeError(
                f'`guild` can be `{Guild.__name__}`, `int`, got {guild.__class__.__name__}; {guild!r}.'
            )
        
        guild = GUILDS.get(guild_id, None)
    
    return guild, guild_id


def get_achievement_id(achievement):
    """
    Gets the achievement identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(
                f'`achievement` can be `{Achievement.__name__}`, `int`, got '
                f'{achievement.__class__.__name__}; {achievement!r}.'
            )
    
    return achievement_id


def get_achievement_and_id(achievement):
    """
    Gets the achievement and it's identifier from the given achievement or of it's identifier.
    
    Parameters
    ----------
    achievement : ``Achievement``, `int`
        The achievement, or it's identifier.
    
    Returns
    -------
    achievement : ``Achievement``, `None`
        The achievement if found.
    achievement_id : `int`
        The achievement's identifier.
    
    Raises
    ------
    TypeError
        If `achievement`'s type is incorrect.
    """
    if isinstance(achievement, Achievement):
        achievement_id = achievement.id
    else:
        achievement_id = maybe_snowflake(achievement)
        if achievement_id is None:
            raise TypeError(
                f'`achievement` can be `{Achievement.__name__}`, `int`, got '
                f'{achievement.__class__.__name__}; {achievement!r}.'
            )
        
        achievement = None
    
    return achievement, achievement_id


def get_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif message is None:
        raise TypeError(
            '`message` was given as `None`. Make sure to call message create methods with non-empty content(s).'
        )
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
    
    return channel_id, message_id


def get_message_and_channel_id_and_message_id(message):
    """
    Gets the message's channel's and it's own identifier.
    
    Parameters
    ----------
    message : ``Message``, `tuple` (`int`, `int`)
        The message or it' representation.
    
    Returns
    -------
    message : `None`, ``Message``
        The message in context if found.
    channel_id : `int`
        The message's channel's identifier.
    message_id : `int`
        The message's identifier.
    
    Raises
    ------
    TypeError
        If `message`'s type is incorrect.
    """
    # 1.: Message
    # 4.: None -> raise
    # 5.: `tuple` (`int`, `int`)
    # 6.: raise
    if isinstance(message, Message):
        channel_id = message.channel_id
        message_id = message.id
    elif message is None:
        raise TypeError(
            '`message` was given as `None`. Make sure to call message create methods with non-empty content(s).'
        )
    else:
        snowflake_pair = maybe_snowflake_pair(message)
        if snowflake_pair is None:
            raise TypeError(
                f'`message` can be `{Message.__name__}`, '
                f'`tuple` of (`int`, `int`), got {message.__class__.__name__}; {message!r}.'
            )
        
        channel_id, message_id = snowflake_pair
        message = MESSAGES.get(message_id, None)
    
    return message, channel_id, message_id


def get_role_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `int`
        The role, or it's identifier.
    
    Returns
    -------
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
    else:
        role_id = maybe_snowflake(role)
        if role_id is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `int`, got {role.__class__.__name__}; {role!r}.'
            )
    
    return role_id


def get_role_role_guild_id_and_id(role):
    """
    Gets the role identifier from the given role or of it's identifier.
    
    Parameters
    ----------
    role : ``Role``, `tuple` (`int`, `int`)
        The role, or a `guild-id`, `role-id` pair.
    
    Returns
    -------
    role : `None`, ``Role``
        The respective role.
    
    guild_id : `int`
        The role's guild's identifier.
    
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        role_id = role.id
        guild_id = role.guild_id
    
    else:
        snowflake_pair = maybe_snowflake_pair(role)
        if snowflake_pair is None:
            raise TypeError(
                f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
            )
        
        guild_id, role_id = snowflake_pair
        role = ROLES.get(role_id, None)
    
    return role, guild_id, role_id


def get_role_guild_id_and_id(role):
    """
    Gets the role's and it's guild's identifier from the given role or of a `guild-id`, `role-id` pair.
    
    Parameters
    ----------
    role : ``Role``, `tuple` (`int`, `int`)
        The role, or a `guild-id`, `role-id` pair.
    
    Returns
    -------
    guild_id : `int`
        The role's guild's identifier.
    
    role_id : `int`
        The role's identifier.
    
    Raises
    ------
    TypeError
        If `role`'s type is incorrect.
    """
    if isinstance(role, Role):
        return role.guild_id, role.id
    
    snowflake_pair = maybe_snowflake_pair(role)
    if snowflake_pair is not None:
        return snowflake_pair
    
    raise TypeError(
        f'`role` can be `{Role.__name__}`, `tuple` (`int`, `int`), got {role.__class__.__name__}; {role!r}.'
    )


def get_webhook_id(webhook):
    """
    Gets the webhook's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        webhook_id = webhook.id
    else:
        webhook_id = maybe_snowflake(webhook)
        if webhook_id is None:
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `int`, got {webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return webhook_id


def get_webhook_and_id(webhook):
    """
    Gets the webhook and it's identifier from the given webhook or of it's identifier.
    
    Parameters
    ----------
    webhook : ``Webhook``, `int`
        The webhook, or it's identifier.
    
    Returns
    -------
    webhook : ``Webhook``, `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            break
        
        webhook_id = maybe_snowflake(webhook)
        if (webhook_id is not None):
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
            
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `int`, got {webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return webhook, webhook_id


def get_webhook_id_and_token(webhook):
    """
    Gets the webhook's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    if isinstance(webhook, Webhook):
        snowflake_token_pair = webhook.id, webhook.token
    else:
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is None):
            raise TypeError(
                f'`webhook` can be `{Webhook.__name__}`, `tuple` (`int`, `str`), got '
                f'{webhook.__class__.__name__}; {webhook!r}.'
            )
    
    return snowflake_token_pair


def get_webhook_and_id_and_token(webhook):
    """
    Gets the webhook, it's identifier and token from the given webhook or it's token, identifier pair.
    
    Parameters
    ----------
    webhook : ``Webhook``, `tuple` (`int`, `str`)
        The webhook or it's id and token.
    
    Returns
    -------
    webhook : ``Webhook``, `None`
        The webhook if any.
    webhook_id : `int`
        The webhook's identifier.
    webhook_token : `str`
        The webhook's token.
    
    Raises
    ------
    TypeError
        If `webhook`'s type is incorrect.
    """
    while True:
        if isinstance(webhook, Webhook):
            webhook_id = webhook.id
            webhook_token = webhook.token
            break
        
        snowflake_token_pair = maybe_snowflake_token_pair(webhook)
        if (snowflake_token_pair is not None):
            webhook_id, webhook_token = snowflake_token_pair
            
            try:
                webhook = USERS[webhook_id]
            except KeyError:
                webhook = None
                break
            
            if isinstance(webhook, Webhook):
                break
        
        raise TypeError(
            f'`webhook` can be `{Webhook.__name__}`, `tuple` (`int`, `str`), got '
            f'{webhook.__class__.__name__}; {webhook!r}.'
        )
    
    return webhook, webhook_id, webhook_token


def get_reaction_emoji_value_and_type(reaction):
    """
    Gets the reaction representation form of the given reaction.
    
    Parameters
    ----------
    reaction : ``Reaction``, ``Emoji``, `str`
        The emoji to get it's reaction form of.
    
    Returns
    -------
    emoji_value : `str`
        The emoji's reaction form.
    reaction_type : ``ReactionType``
        The reaction's type.
    
    Raises
    ------
    TypeError
        If `reaction`'s type is incorrect.
    """
    if isinstance(reaction, Emoji):
        return reaction.as_reaction, ReactionType.standard
    
    if isinstance(reaction, Reaction):
        return reaction.emoji.as_reaction, reaction.type
    
    if isinstance(reaction, str):
        return reaction, ReactionType.standard
    
    raise TypeError(
        f'`reaction` can be `{Reaction.__name__}`, `{Emoji.__name__}`, `str`, '
        f'got {reaction.__class__.__name__}; {reaction!r}.'
    )


def get_emoji_guild_id_and_id(emoji):
    """
    Gets the emoji's and it's guild's identifier from the given emoji.
    
    Parameters
    ----------
    emoji : ``Emoji``, `tuple` (`int`, `int`)
        The emoji, or `guild-id`, `emoji-id` pair.
    
    Returns
    -------
    snowflake_pair : `tuple` (`int`, `int`)
        The emoji's guild's and it's own identifier if applicable.
    
    Raises
    ------
    TypeError
        If `emoji`'s type is incorrect.
    """
    if isinstance(emoji, Emoji):
        snowflake_pair = emoji.guild_id, emoji.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(emoji)
        if snowflake_pair is None:
            raise TypeError(
                f'`emoji` can be `{Emoji.__name__}`, `tuple` (`int`, `int`), got {emoji.__class__.__name__}; '
                f'{emoji!r}.'
            )
    
    return snowflake_pair


def get_sticker_and_id(sticker):
    """
    Gets sticker and it's identifier from the given sticker or of it's identifier.
    
    Parameters
    ----------
    sticker : ``Sticker``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker : ``Sticker``, `None`
        The sticker if found.
    sticker_id : `int`
        The sticker's identifier.
    
    Raises
    ------
    TypeError
        If `sticker`'s type is incorrect.
    """
    while True:
        if isinstance(sticker, Sticker):
            sticker_id = sticker.id
            break
        
        sticker_id = maybe_snowflake(sticker)
        if (sticker_id is not None):
            sticker = STICKERS.get(sticker_id, None)
            break
        
        raise TypeError(
            f'`sticker` can be `{Sticker.__name__}`, `int`, got {sticker.__class__.__name__}; {sticker!r}.'
        )
        
    return sticker, sticker_id


def get_sticker_pack_and_id(sticker_pack):
    """
    Gets sticker pack and it's identifier from the given sticker pack or of it's identifier.
    
    Parameters
    ----------
    sticker_pack : ``StickerPack``, `int`
        The sticker, or it's identifier.
    
    Returns
    -------
    sticker_pack : ``StickerPack``, `None`
        The sticker pack if found.
    sticker_pack_id : `int`
        The sticker pack's identifier.
    
    Raises
    ------
    TypeError
        If `sticker_pack`'s type is incorrect.
    """
    while True:
        if isinstance(sticker_pack, StickerPack):
            sticker_pack_id = sticker_pack.id
            break
        
        sticker_pack_id = maybe_snowflake(sticker_pack)
        if (sticker_pack_id is not None):
            sticker_pack = STICKER_PACKS.get(sticker_pack_id, None)
            break
        
        raise TypeError(
            f'`sticker_pack` can be `{StickerPack.__name__}`, `int`, got {sticker_pack.__class__.__name__}; '
            f'{sticker_pack!r}.'
        )
        
    return sticker_pack, sticker_pack_id


def get_scheduled_event_guild_id_and_id(scheduled_event):
    """
    Gets the scheduled event's and it's identifier from the given scheduled event or from a tuple of 2 identifiers.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `tuple` of (`int`, `int`)
        The scheduled event, or a tuple of two identifiers.
    
    Returns
    -------
    guild_id : `int`
        The scheduled event's guild's identifier.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        return scheduled_event.guild_id, scheduled_event.id
    
    
    snowflake_pair = maybe_snowflake_pair(scheduled_event)
    if snowflake_pair is None:
        raise TypeError(
            f'`scheduled_event` can be `{ScheduledEvent.__name__}`, `tuple` of (`int`, `int`), got '
            f'{scheduled_event.__class__.__name__}; {scheduled_event!r}.'
        )
    
    return snowflake_pair


def get_scheduled_event_and_guild_id_and_id(scheduled_event):
    """
    Gets the scheduled event, it's guild's identifier and it's identifier.
    
    Parameters
    ----------
    scheduled_event : ``ScheduledEvent``, `tuple` (`int`, `int`)
        The scheduled event, or it's identifier.
    
    Returns
    -------
    scheduled_event : `None`, ``ScheduledEvent``
        The scheduled event.
    guild_id : `int`
        The scheduled event's guild's identifier.
    scheduled_event_id : `int`
        The scheduled event's identifier.
    
    Raises
    ------
    TypeError
        If `scheduled_event`'s type is incorrect.
    """
    if isinstance(scheduled_event, ScheduledEvent):
        scheduled_event =  scheduled_event
        scheduled_event_id = scheduled_event.id
        guild_id = scheduled_event.guild_id
    else:
        snowflake_pair = maybe_snowflake_pair(scheduled_event)
        if snowflake_pair is None:
            raise TypeError(
                f'`scheduled_event` can be `{ScheduledEvent.__name__}`, `tuple` of (`int`, `int`), got '
                f'{scheduled_event.__class__.__name__}; {scheduled_event!r}.'
            )
        
        guild_id, scheduled_event_id = snowflake_pair
        
        scheduled_event = SCHEDULED_EVENTS.get(scheduled_event_id, None)
    
    return scheduled_event, guild_id, scheduled_event_id


def get_application_command_id(application_command):
    """
    Gets the application command's identifier.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
    
    return application_command_id


def get_application_command_and_id(application_command):
    """
    Gets the application command's identifier.
    
    Parameters
    ----------
    application_command : ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command : `None`, ``ApplicationCommand``
        The application command if exists.
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
        
        application_command = APPLICATION_COMMANDS.get(application_command_id, None)
    
    return application_command, application_command_id


def get_application_command_id_nullable(application_command):
    """
    Gets the application command's identifier.
    
    > The application command can be `None`. At that case `application_command_id` will default to `0`.
    
    Parameters
    ----------
    application_command : `None`, ``ApplicationCommand``, `int`
        The application command or it's identifier.
    
    Returns
    -------
    application_command_id : `int`
        The application command's identifier.
    
    Raises
    ------
    TypeError
        If `application_command`'s type is incorrect.
    """
    if application_command is None:
        application_command_id = 0
    
    elif isinstance(application_command, ApplicationCommand):
        application_command_id = application_command.id
    
    else:
        application_command_id = maybe_snowflake(application_command)
        if application_command_id is None:
            raise TypeError(
                f'`application_command` can be `{ApplicationCommand.__name__}`, `int`, got '
                f'{application_command.__class__.__name__}; {application_command!r}.'
            )
    
    return application_command_id


def get_auto_moderation_rule_guild_id_and_id(auto_moderation_rule):
    """
    Gets the auto moderation rule's guild's identifier and it's own identifier.
    
    Parameters
    ----------
    auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
        The auto moderation rule, or it's identifier as a `guild-id`, `rule-id` pair..
    
    Returns
    -------
    guild_id : `int`
        The auto moderation rule's guild's identifier.
    auto_moderation_rule_id : `int`
        The auto moderation rule's identifier.
    
    Raises
    ------
    TypeError
        If `auto_moderation_rule`'s type is incorrect.
    """
    if isinstance(auto_moderation_rule, AutoModerationRule):
        guild_id = auto_moderation_rule.guild_id
        rule_id = auto_moderation_rule.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(auto_moderation_rule)
        if snowflake_pair is None:
            raise TypeError(
                f'`auto_moderation_rule` can be `{AutoModerationRule.__name__}`, `tuple` of (`int`, `int`), got '
                f'{auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        guild_id, rule_id = snowflake_pair
    
    return guild_id, rule_id


def get_auto_moderation_rule_and_guild_id_and_id(auto_moderation_rule):
    """
    Gets the auto moderation, its guild's identifier and it's own identifier.
    
    Parameters
    ----------
    auto_moderation_rule : ``AutoModerationRule``, `tuple` (`int`, `int`)
        The auto moderation rule, or it's identifier as a `guild-id`, `rule-id` pair..
    
    Returns
    -------
    auto_moderation_rule : `None`, ``AutoModerationRule``
        The auto moderation rule.
    guild_id : `int`
        The auto moderation rule's guild's identifier.
    auto_moderation_rule_id : `int`
        The auto moderation rule's identifier.
    
    Raises
    ------
    TypeError
        If `auto_moderation_rule`'s type is incorrect.
    """
    if isinstance(auto_moderation_rule, AutoModerationRule):
        guild_id = auto_moderation_rule.guild_id
        rule_id = auto_moderation_rule.id
    
    else:
        snowflake_pair = maybe_snowflake_pair(auto_moderation_rule)
        if snowflake_pair is None:
            raise TypeError(
                f'`auto_moderation_rule` can be `{AutoModerationRule.__name__}`, `tuple` of (`int`, `int`), got '
                f'{auto_moderation_rule.__class__.__name__}; {auto_moderation_rule!r}.'
            )
        
        guild_id, rule_id = snowflake_pair
        auto_moderation_rule = AUTO_MODERATION_RULES.get(rule_id, None)
    
    return auto_moderation_rule, guild_id, rule_id


def get_forum_tag_id(forum_tag):
    """
    Gets the forum tag's identifier.
    
    Parameters
    ----------
    forum_tag : ``ForumTag``, `int`
        The forum tag or its identifier.
    
    Returns
    -------
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Raises
    ------
    TypeError
        - If `forum_tag` type is incorrect.
    """
    if isinstance(forum_tag, ForumTag):
        forum_tag_id = forum_tag.id
    
    else:
        forum_tag_id = maybe_snowflake(forum_tag)
        if (forum_tag_id is None):
            raise TypeError(
                f'`forum_tag` can be `{ForumTag.__name__}, `int`, got {forum_tag.__class__.__name__}; {forum_tag!r}.'
            )
    
    return forum_tag_id


def get_forum_tag_and_id(forum_tag):
    """
    Gets the forum tag and it's identifier.
    
    Parameters
    ----------
    forum_tag : ``ForumTag``, `int`
        The forum tag or its identifier.
    
    Returns
    -------
    forum_tag : `None`, ``ForumTag``
        The forum tag if exists.
    forum_tag_id : `int`
        The forum tag's identifier.
    
    Raises
    ------
    TypeError
        - If `forum_tag` type is incorrect.
    """
    if isinstance(forum_tag, ForumTag):
        forum_tag_id = forum_tag.id
    
    else:
        forum_tag_id = maybe_snowflake(forum_tag)
        if (forum_tag_id is None):
            raise TypeError(
                f'`forum_tag` can be `{ForumTag.__name__}, `int`, got {forum_tag.__class__.__name__}; {forum_tag!r}.'
            )
        
        forum_tag = FORUM_TAGS.get(forum_tag_id, None)
    
    return forum_tag, forum_tag_id


def get_permission_overwrite_target_id(permission_overwrite):
    """
    Returns the given `permission_overwrite`'s target id.
    
    Parameters
    ----------
    permission_overwrite : ``PermissionOverwrite``
        The permission overwrite or it's target id.
    
    Returns
    -------
    target_id : `int`
    
    Raises
    ------
    TypeError
        - If `permission_overwrite` type is incorrect.
    """
    if isinstance(permission_overwrite, PermissionOverwrite):
        target_id = permission_overwrite.target_id
    
    else:
        target_id = maybe_snowflake(permission_overwrite)
        if (target_id is None):
            raise TypeError(
                f'`permission_overwrite` can be `{PermissionOverwrite.__name__}`, `int` got '
                f'{permission_overwrite.__class__.__name__}; {permission_overwrite!r}.'
            )
    
    return target_id


def _check_required_scope(access, required_scope):
    if (required_scope is not None) and (not access.has_scope(required_scope)):
        raise ValueError(
            f'The given `access` not grants {required_scope.name!r} scope, what is required, '
            f'got {access!r}.'
        )


def get_oauth2_access_token(access, required_scope = None):
    """
    Returns the given oauth2 access's access token. Accepts both access and just token as well.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``, `str`
        Oauth2 access to the respective user or it's access token.
    required_scope : `None`, ``Oauth2Scope`` = `None`, Optional
        Required scope of the access if any.
    
    Returns
    -------
    access_token : `str`
    
    Raises
    ------
    TypeError
        - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
    ValueError
        - If the given `access` is not providing the required scope.
    """
    if isinstance(access, (Oauth2Access, Oauth2User)):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
    
    elif isinstance(access, str):
        access_token = access
    
    else:
        raise TypeError(
            f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`'
            f', got {access.__class__.__name__}; {access!r}.'
        )
    
    return access_token


def get_oauth2_access_token_and_user_id(access, user, required_scope = None):
    """
    Returns the given oauth2 access's access token and the user's identifier.
    
    Parameters
    ----------
    access : ``Oauth2Access``, ``Oauth2User``, `str`
        Oauth2 access to the respective user or it's access token.
    required_scope : `None`, ``Oauth2Scope`` = `None`, Optional
        Required scope of the access if any.
    
    Raises
    ------
    TypeError
        - If `access` is not ``Oauth2Access``, ``Oauth2User``, `str`.
        - If `user` is not `None`, ``ClientUserBase``.
        - If `user.id` could nto be determined.
    ValueError
        - If the given `access` is not providing the required scope.
        - If `user` and `access` refers to a different user.
    """
    user_id = get_user_id_nullable(user)
    
    if isinstance(access, Oauth2Access):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
    
    elif isinstance(access, Oauth2User):
        _check_required_scope(access, required_scope)
        access_token = access.access_token
        
        if user_id and (user_id != access.id):
            raise ValueError(
                f'The given `user` and `access` refers to different users, got user = {user!r}, '
                f'access = {access!r}.'
            )
        
        user_id = access.id
    
    elif isinstance(access, str):
        access_token = access
    
    else:
        raise TypeError(
            f'`access` can be `{Oauth2Access.__name__}`, `{Oauth2User.__name__}`, `str`, got '
            f'{access.__class__.__name__}; {access!r}.'
        )
    
    
    if not user_id:
        raise TypeError(
            f'`user` was not detectable neither from `user` nor from `access` parameters, got '
            f'user = {user!r}, access = {access!r}.'
        )
    
    return access_token, user_id


def get_soundboard_sound_and_guild_id_and_id(soundboard_sound):
    """
    Gets the soundboard sound identifier from the given soundboard sound or of it's identifier.
    
    Parameters
    ----------
    soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
        The soundboard sound, or a `guild-id`, `soundboard-sound-id` pair.
    
    Returns
    -------
    soundboard_sound : `None`, ``SoundboardSound``
        The respective soundboard sound.
    
    guild_id : `int`
        The soundboard sound's guild's identifier.
    
    soundboard_sound_id : `int`
        The soundboard sound's identifier.
    
    Raises
    ------
    TypeError
        If `soundboard_sound`'s type is incorrect.
    """
    if isinstance(soundboard_sound, SoundboardSound):
        soundboard_sound_id = soundboard_sound.id
        guild_id = soundboard_sound.guild_id
    
    else:
        snowflake_pair = maybe_snowflake_pair(soundboard_sound)
        if snowflake_pair is None:
            raise TypeError(
                f'`soundboard_sound` can be `{SoundboardSound.__name__}`, `tuple` (`int`, `int`), '
                f'got {soundboard_sound.__class__.__name__}; {soundboard_sound!r}.'
            )
        
        guild_id, soundboard_sound_id = snowflake_pair
        soundboard_sound = SOUNDBOARD_SOUNDS.get(soundboard_sound_id, None)
    
    return soundboard_sound, guild_id, soundboard_sound_id


def get_soundboard_sound_guild_id_and_id(soundboard_sound):
    """
    Gets the soundboard_sound's and it's guild's identifier from the given soundboard_sound or of a `guild-id`,
    `soundboard_sound-id` pair.
    
    Parameters
    ----------
    soundboard_sound : ``SoundboardSound``, `tuple` (`int`, `int`)
        The soundboard sound, or a `guild-id`, `soundboard-sound-id` pair.
    
    Returns
    -------
    guild_id : `int`
        The soundboard sound's guild's identifier.
    
    soundboard_sound_id : `int`
        The soundboard sound's identifier.
    
    Raises
    ------
    TypeError
        If `soundboard_sound`'s type is incorrect.
    """
    if isinstance(soundboard_sound, SoundboardSound):
        return soundboard_sound.guild_id, soundboard_sound.id
    
    snowflake_pair = maybe_snowflake_pair(soundboard_sound)
    if snowflake_pair is not None:
        return snowflake_pair
    
    raise TypeError(
        f'`soundboard_sound` can be `{SoundboardSound.__name__}`, `tuple` (`int`, `int`), '
        f'got {soundboard_sound.__class__.__name__}; {soundboard_sound!r}.'
    )


def get_entitlement_id(entitlement):
    """
    Gets the entitlement's identifier.
    
    Parameters
    ----------
    entitlement : ``Entitlement``, `int`
        The entitlement or its identifier.
    
    Returns
    -------
    entitlement_id : `int`
        The entitlement's identifier.
    
    Raises
    ------
    TypeError
        - If `entitlement` type is incorrect.
    """
    if isinstance(entitlement, Entitlement):
        entitlement_id = entitlement.id
    
    else:
        entitlement_id = maybe_snowflake(entitlement)
        if (entitlement_id is None):
            raise TypeError(
                f'`entitlement` can be `{Entitlement.__name__}, `int`, '
                f'got {entitlement.__class__.__name__}; {entitlement!r}.'
            )
    
    return entitlement_id


def get_poll_answer_and_id(poll_answer):
    """
    Gets the poll answer and its identifier.
    
    Parameters
    ----------
    poll_answer : ``PollAnswer``, `int`
        The poll answer to get its identifier.
    
    Returns
    -------
    poll_answer : `None`, ``PollAnswer``
    poll_answer_id : `int`
    
    Raises
    ------
    TypeError
        - If `poll_answer` type is incorrect.
    """
    if isinstance(poll_answer, PollAnswer):
        return poll_answer, poll_answer.id
    
    poll_answer_id = maybe_snowflake(poll_answer)
    if poll_answer_id is None:
        raise TypeError(
            f'`poll_answer` can be `{PollAnswer.__name__}, `int`, '
            f'got {poll_answer.__class__.__name__}; {poll_answer!r}.'
        )
    
    return None, poll_answer_id
